# -*- coding: utf-8 -*-

"""
Data Table View
===============

"""


# %% IMPORTS
# Built-in imports

# Package imports
from qtpy import QtCore as QC, QtGui as QG, QtWidgets as QW

# GuiPy imports
from guipy import layouts as GL, widgets as GW
from guipy.plugins.data_table.widgets.headers import (
    HorizontalHeaderView, VerticalHeaderView)
from guipy.plugins.data_table.widgets.model import DataTableModel, to_base_26
from guipy.plugins.data_table.widgets.selection_model import (
    DataTableSelectionModel)
from guipy.widgets import get_box_value, get_modified_signal, set_box_value

# All declaration
__all__ = ['DataTableView']


# %% CLASS DEFINITIONS
# Define table view widget for the DataTable plugin
# TODO: Save all data in the widget before resizing
# HINT: https://doc.qt.io/qt-5/model-view-programming.html
class DataTableView(GW.QTableView):
    # Initialize DataTableView class
    def __init__(self, parent=None, *args, **kwargs):
        # Call super constructor
        super().__init__(parent)

        # Set up the data table widget
        self.init(*args, **kwargs)

    # This function sets up the data table widget
    def init(self, import_func=None):
        # Set model for the data table widget
        self.setModel(DataTableModel(self, import_func))

        # Create selection model for the data table widget
        selection_model = DataTableSelectionModel(self.model(), self)

        # Set selection model for the data table widget
        self.selectionModel().deleteLater()
        self.setSelectionModel(selection_model)

        # Create headers
        self.create_headers()

    # Override closeEvent to do automatic clean-up
    def closeEvent(self, *args, **kwargs):
        # Delete the model
        self.model().delete()

        # Call super event
        super().closeEvent(*args, **kwargs)

    # This function returns the number of columns in the model
    def columnCount(self):
        return(self.model().columnCount())

    # This function sets the number of columns in the model
    def setColumnCount(self, count):
        # Calculate the difference between count and columnCount
        diff = count-self.columnCount()

        # If diff is negative, remove columns from the end
        if(diff < 0):
            self.model().removeColumns(count=abs(diff))
        # If diff is positive, append columns to the end
        elif(diff > 0):
            self.model().insertColumns(count=abs(diff))
        # If diff is zero, do nothing
        else:
            pass

    # This function returns the number of rows in the model
    def rowCount(self):
        return(self.model().rowCount())

    # This function sets the number of rows in the model
    def setRowCount(self, count):
        # Calculate the difference between count and rowCount
        diff = count-self.rowCount()

        # Extend or shorten the data table model accordingly
        if(diff < 0):
            self.model().removeRows(count=abs(diff))
        elif(diff > 0):
            self.model().insertRows(count=abs(diff))
        else:
            pass

    # This function creates the headers for the data table
    def create_headers(self):
        # Create context menus for the horizontal and vertical headers
        self.create_header_context_menus()

        # Create popup editor for the horizontal header
        self.h_header_editor = HorizontalHeaderPopup(self)

        # Create horizontal and vertical headers for the data table widget
        self.h_header = HorizontalHeaderView(
            parent=self,
            context_menu=self.show_horizontal_header_context_menu,
            double_clicked=self.h_header_editor)
        self.v_header = VerticalHeaderView(
            parent=self,
            context_menu=self.show_vertical_header_context_menu)

        # Set headers
        self.setHorizontalHeader(self.h_header)
        self.setVerticalHeader(self.v_header)

    # This function creates the header context menus
    def create_header_context_menus(self):
        self.create_horizontal_header_context_menu()
        self.create_vertical_header_context_menu()

    # This function creates the horizontal header context menu
    def create_horizontal_header_context_menu(self):
        # Create context menu
        menu = GW.QMenu('H_Header', parent=self)

        # Add insert_above action to menu
        insert_above_act = GW.QAction(
            self, "Insert column left",
            statustip="Insert a new column left of this one",
            triggered=self.insert_cols)
        menu.addAction(insert_above_act)

        # Add insert_below action to menu
        insert_below_act = GW.QAction(
            self, "Insert column right",
            statustip="Insert a new column right of this one",
            triggered=self.insert_cols_after)
        menu.addAction(insert_below_act)

        # Add remove action to menu
        remove_act = GW.QAction(
            self, "Remove column",
            statustip="Remove this column",
            triggered=self.remove_cols)
        menu.addAction(remove_act)

        # Add clear action to menu
        clear_act = GW.QAction(
            self, "Clear column",
            statustip="Clear this column",
            triggered=self.clear_cols)
        menu.addAction(clear_act)

        # Add hide action to menu
        hide_act = GW.QAction(
            self, "Hide column",
            statustip="Hide this column",
            triggered=self.hide_cols)
#        menu.addAction(hide_act)

        # Set last requested col to 0
        self._last_context_col = 0

        # Save made menu as an attribute
        self.h_header_menu = menu

    # This function creates the vertical header context menu
    def create_vertical_header_context_menu(self):
        # Create context menu
        menu = GW.QMenu('V_Header', parent=self)

        # Add insert_above action to menu
        insert_above_act = GW.QAction(
            self, "Insert row above",
            statustip="Insert a new row above this one",
            triggered=self.insert_rows)
        menu.addAction(insert_above_act)

        # Add insert_below action to menu
        insert_below_act = GW.QAction(
            self, "Insert row below",
            statustip="Insert a new row below this one",
            triggered=self.insert_rows_after)
        menu.addAction(insert_below_act)

        # Add remove action to menu
        remove_act = GW.QAction(
            self, "Remove row",
            statustip="Remove this row",
            triggered=self.remove_rows)
        menu.addAction(remove_act)

        # Add clear action to menu
        clear_act = GW.QAction(
            self, "Clear row",
            statustip="Clear this row",
            triggered=self.clear_rows)
        menu.addAction(clear_act)

        # Add hide action to menu
        hide_act = GW.QAction(
            self, "Hide row",
            statustip="Hide this row",
            triggered=self.hide_rows)
#        menu.addAction(hide_act)

        # Set last requested row to 0
        self._last_context_row = 0

        # Save made menu as an attribute
        self.v_header_menu = menu

    # This function shows the horizontal header context menu
    # TODO: Figure out how to use the visual index for insertions
    @QC.Slot(QC.QPoint)
    def show_horizontal_header_context_menu(self, pos):
        # Obtain the logical column the context menu was requested for
        logical_index = self.h_header.logicalIndexAt(pos)

        # Determine which visual column this is
        visual_index = self.h_header.visualIndex(logical_index)

        # Save both indexes
#        self._last_context_col = (logical_index, visual_index)
        self._last_context_col = logical_index

        # Show context menu
        self.h_header_menu.popup(QG.QCursor.pos())

    # This function shows the vertical header context menu
    @QC.Slot(QC.QPoint)
    def show_vertical_header_context_menu(self, pos):
        # Obtain the logical row the context menu was requested for
        logical_index = self.v_header.logicalIndexAt(pos)

        # Determine which visual row this is
        visual_index = self.v_header.visualIndex(logical_index)

        # Save both indexes
#        self._last_context_row = (logical_index, visual_index)
        self._last_context_row = logical_index

        # Show context menu
        self.v_header_menu.popup(QG.QCursor.pos())

    # This function inserts columns into the data table before given column
    @QC.Slot()
    @QC.Slot(int)
    def insert_cols(self, n_cols=1):
        self.model().insertColumns(self._last_context_col, n_cols)

    # This function inserts columns into the data table after given column
    @QC.Slot()
    @QC.Slot(int)
    def insert_cols_after(self, n_cols=1):
        self.model().insertColumns(self._last_context_col+n_cols, n_cols)

    # This function removes a given column in the data table
    @QC.Slot()
    @QC.Slot(int)
    def remove_cols(self, n_cols=1):
        self.model().removeColumns(self._last_context_col, n_cols)

    # This function clears a given column in the data table
    @QC.Slot()
    @QC.Slot(int)
    def clear_cols(self, n_cols=1):
        self.model().clearColumns(self._last_context_col, n_cols)

    # This function hides a given column in the data table
    @QC.Slot()
    @QC.Slot(int)
    def hide_cols(self, n_cols=1):
        self.model().hideColumns(self._last_context_col, n_cols)

    # This function inserts rows into the data table before given row
    @QC.Slot()
    @QC.Slot(int)
    def insert_rows(self, n_rows=1):
        self.model().insertRows(self._last_context_row, n_rows)

    # This function inserts rows into the data table after given row
    @QC.Slot()
    @QC.Slot(int)
    def insert_rows_after(self, n_rows=1):
        self.model().insertRows(self._last_context_row+n_rows, n_rows)

    # This function removes a given row in the data table
    @QC.Slot()
    @QC.Slot(int)
    def remove_rows(self, n_rows=1):
        self.model().removeRows(self._last_context_row, n_rows)

    # This function clears a given row in the data table
    @QC.Slot()
    @QC.Slot(int)
    def clear_rows(self, n_rows=1):
        self.model().clearRows(self._last_context_row, n_rows)

    # This function hides a given row in the data table
    @QC.Slot()
    @QC.Slot(int)
    def hide_rows(self, n_rows=1):
        self.model().hideRows(self._last_context_row, n_rows)


# Define class for showing a popup editor for the horizontal header
class HorizontalHeaderPopup(GW.QDialog):
    # Initialize HorizontalHeaderPopup class
    def __init__(self, data_table_view_obj, *args, **kwargs):
        # Save provided DataTableView object
        self.data_table = data_table_view_obj
        self.model = self.data_table.model()

        # Call super constructor
        super().__init__(data_table_view_obj)

        # Set up the header editor
        self.init(*args, **kwargs)

    # This function shows the editor
    @QC.Slot(int)
    def __call__(self, col):
        # Save which column index was requested
        self.col = col

        # Get the column that was requested
        column = self.model.dataColumn(col)

        # Get the dtype of this column
        dtype = self.model.dtypes[column.dtype.type]

        # Determine the names of all other columns
        used_column_names = set(self.model.columnNames())
        used_column_names.difference_update(['', column.name])
        self.used_column_names = used_column_names

        # Set the base name, name and dtype of this column
        base_name = "Column %s" % (to_base_26(col+1))
        set_box_value(self.base_name_label, base_name)
        set_box_value(self.n_val_box, column.count())
        set_box_value(self.name_box, column.name)
        set_box_value(self.dtype_box, dtype)

        # Set keyboard focus to the name_box and select it
        self.name_box.setFocus(True)
        self.name_box.selectAll()

        # Show the popup
        self.show()

    # This function sets up the horizontal header popup editor
    # TODO: Allow for a simple formula to be given for a column?
    def init(self):
        # Install event filter to catch events that should close the popup
        self.installEventFilter(self)

        # Set dialog flags
        self.setWindowFlags(
            QC.Qt.Popup |
            QC.Qt.FramelessWindowHint)

        # Create a form layout
        layout = GL.QFormLayout(self)

        # Add a label stating the base name of the column
        self.base_name_label = GW.QLabel("")
        self.base_name_label.setAlignment(QC.Qt.AlignCenter)
        layout.addRow(self.base_name_label)

        # Create a n_val label
        n_val_box = GW.QLabel()
        n_val_box.setToolTip("Number of values in this column")

        # Add it to the layout
        layout.addRow("# of values", n_val_box)
        self.n_val_box = n_val_box

        # Create a name line-edit
        name_box = GW.QLineEdit()
        name_box.setToolTip("Set a custom name for this column or leave empty "
                            "to use its default name")
        get_modified_signal(name_box).connect(self.column_name_changed)

        # Add it to the layout
        layout.addRow("Name", name_box)
        self.name_box = name_box

        # Create a dtype combobox
        dtype_box = GW.QComboBox()
        dtype_box.setToolTip("Set the data type for this column")
        dtype_box.addItems(self.model.dtypes.values())
        dtype_box.popup_hidden.connect(lambda: name_box.setFocus(True))

        # Add it to the layout
        layout.addRow("Data type", dtype_box)
        self.dtype_box = dtype_box

    # Override eventFilter to filter out clicks, ESC and Enter
    def eventFilter(self, widget, event):
        # Check if the event involves anything for which the popup should close
        if (((event.type() == QC.QEvent.MouseButtonPress) and
             not self.geometry().contains(event.globalPos())) or
            ((event.type() == QC.QEvent.KeyPress) and
             event.key() in (QC.Qt.Key_Escape,
                             QC.Qt.Key_Enter,
                             QC.Qt.Key_Return))):
            # Exit the editor
            self.hide()
            self.name_box.setFocus(False)
            return(True)

        # Else, process events as normal
        else:
            return(super().eventFilter(widget, event))

    # Override hideEvent to automatically update the header
    def hideEvent(self, *args, **kwargs):
        # Set the column name and dtype
        self.set_column_name(get_box_value(self.name_box))
        self.set_column_dtype(get_box_value(self.dtype_box))

        # Tell data table to update the header of the requested column
        self.data_table.h_header.headerDataChanged(QC.Qt.Horizontal,
                                                   self.col, self.col)

        # Call super event
        super().hideEvent(*args, **kwargs)

    # This function is called whenever the column name is changed
    @QC.Slot(str)
    def column_name_changed(self, name):
        # Determine the position of the tooltip
        pos = self.name_box.mapToGlobal(QC.QPoint(self.name_box.rect().left(),
                                                  self.name_box.height()//2))

        # If this name is not valid, show tooltip with error
        if not self.check_column_name(name):
            err_msg = "This name is either invalid or already taken!"
            GW.QToolTip.showText(pos, err_msg, self.name_box)

        # If it is valid, disable any previous tooltip
        else:
            GW.QToolTip.hideText()

    # This function checks if a given name could be used for a given column
    # TODO: Should I use a QG.QValidator for this?
    @QC.Slot(str)
    def check_column_name(self, name):
        # If name is empty, it is always valid
        if not name:
            return(True)

        # Check if name consists out of one or two capital letters
        # TODO: This assumes a maximum of 702 columns, decreasing modularity
        elif name.isalpha() and name.isupper() and (len(name) <= 2):
            # If so, it is invalid as these are default column names
            return(False)

        # Check if the name is already used for something else
        else:
            return(name not in self.used_column_names)

    # This function is called when the column name is being set
    @QC.Slot(str)
    def set_column_name(self, name):
        # Set the column name if valid
        if self.check_column_name(name):
            self.model.setColumnName(self.col, name)

    # This function is called when the column dtype is being set
    @QC.Slot(str)
    def set_column_dtype(self, dtype):
        # Set the column dtype
        self.model.setColumnDataType(self.col, dtype)



# -*- coding: utf-8 -*-

"""
EntriesBoxes
============

"""


# %% IMPORTS
# Built-in imports


# Package imports
from qtpy import QtCore as QC, QtGui as QG, QtWidgets as QW
from sortedcontainers import SortedDict as sdict, SortedSet as sset

# GuiPy imports
from guipy import layouts as GL, widgets as GW
from guipy.widgets import get_box_value, get_modified_signal, set_box_value

# All declaration
__all__ = ['EditableEntriesBox', 'EntriesBox']


# %% HELPER DEFINITIONS
# Make validator class for the name box in an entries box
class EntryNameBoxValidator(QG.QValidator):
    # Initialize the EntryNameBoxValidator class
    def __init__(self, combobox_obj, entries_box_obj, misc_flag, parent=None):
        """
        Initialize an instance of the :class:`~EntryNameBoxValidator` class.

        Parameters
        ----------
        combobox_obj : :obj:`~PyQt5.QtWidgets.ComboBox` object
            Combobox object for which the editable line must be validated.
        entries_box_obj : :obj:`~EntriesBox` object
            Entries box object for which this validator must be created.
        misc_flag : bool
            Whether or not all non-handled cases should be acceptable or
            invalid.

        Optional
        --------
        parent : :obj:`~PyQt5.QtCore.QObject` object or None. Default: None
            The parent object to use for this validator or *None* for no
            parent.

        """

        # Save the provided entries_box_obj
        self.entries_box_obj = entries_box_obj
        self.combobox_obj = combobox_obj
        self.completer = combobox_obj.completer()

        # Set misc_flag
        if misc_flag:
            self.misc = self.Acceptable
        else:
            self.misc = self.Invalid

        # Call super constructor
        super().__init__(parent)

    # Override validate to reject banned and duplicate entries
    def validate(self, string, pos):
        # Check if string is empty
        if not string:
            # If so, it is acceptable
            state = self.Intermediate

            # Return state
            return(state, string, pos)

        # Create empty list of all current names in the entries box
        cur_names = []

        # Obtain the names of all current entries in the entries box
        for i in range(4, 4+3*self.entries_box_obj.entryCount(), 3):
            name_box = self.entries_box_obj.entries_grid.itemAt(i).widget()
            cur_names.append(get_box_value(name_box))

        # Remove string from it
        try:
            cur_names.remove(string)
        except ValueError:
            pass

        # Check if string can be found in cur_names
        if string in cur_names:
            # If so, it is intermediate
            state = self.Intermediate

            # Return state
            return(state, string, pos)

        # Check if string is already in the completions list
        index = self.completer.completionModel().index(0, 0)
        match = self.completer.completionModel().match(
            index, QC.Qt.EditRole, string, flags=QC.Qt.MatchExactly)

        # Check if there is a match
        if not match:
            # If not, set the completion prefix in the completer
            self.completer.setCompletionPrefix(string)
        else:
            # If so, it is acceptable
            state = self.Acceptable

            # Return state
            return(state, string, pos)

        # Obtain the current completion string
        completion = self.completer.currentCompletion()

        # Check if this completion string is not empty
        if completion:
            # If so, it is acceptable
            state = self.Intermediate

        # Else, check if the string is banned
        elif string in self.entries_box_obj.banned_names:
            # If so, it is intermediate as well
            state = self.Intermediate

        # Else, it is acceptable
        else:
            state = self.misc

        # Return state
        return(state, string, pos)

    # Override fixup to set the entry name to "" when it is invalid
    def fixup(self, string):
        set_box_value(self.combobox_obj, "")


# %% CLASS DEFINITIONS
# Make class for creating entry lists
class EntriesBox(GW.BaseBox):
    """
    Defines the :class:`~EntriesBox` class.

    This widget allows for a series of 'list' entries to be maintained within a
    single box, and is basically a Qt-version of a Python dict.
    Entries that must use a specific widget can be added using the
    :meth:`~addEntryTypes` method, whereas disallowed entry names can be added
    with :meth:`~addBannedNames`.

    """

    # Signals
    modified = QC.Signal([], [dict])

    # Initialize the EntriesBox class
    def __init__(self, parent=None):
        """
        Initialize an instance of the :class:`~EntriesBox` class.

        Optional
        --------
        parent : :obj:`~PyQt5.QtWidgets.QWidget` object or None. Default: None
            The parent widget to use for this entries box or *None* for no
            parent.

        """

        # Call super constructor
        super().__init__(parent, auto_connect=False)

        # Create the entries box
        self.init()

    # This property returns the default 'modified' signal
    @property
    def default_modified_signal(self):
        return(self.modified[dict])

    # This property returns the number of entries in the box
    def entryCount(self):
        return((self.entries_grid.count()//3)-1)

    # This function creates the entries box
    def init(self):
        """
        Sets up the entries box after it has been initialized.

        """

        # Set the height of a single entry
        self.entry_height = 24

        # Create empty dict of non-generic entry types
        self.entry_types = {}

        # Create empty dict of entry defaults
        self.entry_defaults = {}

        # Create empty set of banned entry names
        self.banned_names = sset()

        # Create the box_layout
        box_layout = GL.QVBoxLayout(self)
        box_layout.setContentsMargins(0, 0, 0, 0)

        # Create the entries_grid
        entries_grid = GL.QGridLayout()
        entries_grid.setContentsMargins(0, 0, 0, 0)
        entries_grid.setColumnStretch(1, 1)
        entries_grid.setColumnStretch(2, 2)
        box_layout.addLayout(entries_grid)
        self.entries_grid = entries_grid

        # Add a header
        entries_grid.addWidget(GW.QLabel(""), 0, 0)
        entries_grid.addWidget(GW.QLabel("Entry name"), 0, 1)
        entries_grid.addWidget(GW.QLabel("Entry value"), 0, 2)

        # Add an 'Add'-button
        add_but = GW.QToolButton()
        add_but.setFixedSize(self.entry_height, self.entry_height)
        add_but.setToolTip("Add new entry")
        get_modified_signal(add_but).connect(self.add_entry)
        box_layout.addWidget(add_but)

        # If this theme has an 'add' icon, use it
        if QG.QIcon.hasThemeIcon('add'):
            add_but.setIcon(QG.QIcon.fromTheme('add'))
        # Else, use a simple plus
        else:
            add_but.setText('+')

        # Set size policy
        self.setSizePolicy(QW.QSizePolicy.Preferred, QW.QSizePolicy.Fixed)

        # Set a minimum width for the first column
        self.entries_grid.setColumnMinimumWidth(0, self.entry_height)

    # This function is automatically called whenever 'modified' is emitted
    @QC.Slot()
    def modified_signal_slot(self):
        self.modified[dict].emit(EntriesBox.get_box_value(self))

    # This function returns the proper combobox to be used for name entries
    def get_entry_name_box(self):
        # Create name_box
        name_box = GW.EditableComboBox()

        # Initialize and apply special validator
        validator = EntryNameBoxValidator(name_box, self, False)
        name_box.setValidator(validator)

        # Return name_box
        return(name_box)

    # This function returns whether a given name_box has a valid value
    def is_valid(self, name_box):
        # Determine the current value of the name_box
        entry_name = get_box_value(name_box)

        # Check if the name_box is editable
        if name_box.isEditable():
            # If so, check if its current input is valid
            valid = name_box.lineEdit().hasAcceptableInput()
        else:
            # If not, check if the current input is not empty
            valid = bool(entry_name)

        # Return valid
        return(valid)

    # This function is called whenever a new entry is added
    @QC.Slot()
    def add_entry(self):
        """
        Adds a new entry to the entries box.

        """

        # Create a combobox with the name of the entry
        name_box = self.get_entry_name_box()
        name_box.setToolTip("Select or type name for this entry")
        name_box.addItems(self.entry_types.keys())
        set_box_value(name_box, -1)
        get_modified_signal(name_box).connect(
            lambda: self.create_value_box(name_box))

        # Create a 'Delete'-button
        del_but = GW.QToolButton()
        del_but.setFixedSize(self.entry_height, self.entry_height)
        del_but.setToolTip("Delete this entry")
        get_modified_signal(del_but).connect(
            lambda: self.remove_entry(name_box))

        # If this theme has a 'remove' icon, use it
        if QG.QIcon.hasThemeIcon('remove'):
            del_but.setIcon(QG.QIcon.fromTheme('remove'))
        # Else, use a standard icon
        else:
            del_but.setIcon(del_but.style().standardIcon(
                QW.QStyle.SP_DialogCloseButton))

        # Add a new row to the grid layout
        next_row = self.entries_grid.getItemPosition(
            self.entries_grid.count()-1)[0]+1
        self.entries_grid.addWidget(del_but, next_row, 0)
        self.entries_grid.addWidget(name_box, next_row, 1)
        self.entries_grid.addWidget(GW.QLabel(), next_row, 2)

    # This function is called whenever an entry must be removed
    @QC.Slot(GW.QComboBox)
    def remove_entry(self, name_box):
        """
        Removes the entry associated with the provided `name_box` from the
        entries box.

        """

        # Determine if name_box is valid
        valid = self.is_valid(name_box)

        # Determine at what index the provided name_box currently is in grid
        index = self.entries_grid.indexOf(name_box)

        # As every entry contains 3 items, remove item 3 times at index-1
        for _ in range(3):
            # Remove the current item at index-1
            item = self.entries_grid.takeAt(index-1)

            # Close the widget in this item and delete it
            item.widget().close()
            del item

        # Emit modified signal if name_box was valid
        if valid:
            self.modified.emit()

    # This function is called whenever a new entry field box is requested
    @QC.Slot(GW.QComboBox)
    def create_value_box(self, name_box):
        """
        Creates a value box for the provided `name_box` and its corresponding
        current value, and replaces the current value box with it.

        If the current value of the given `name_box` does not have a specific
        value box defined, a :class:`~guipy.widgets.LongGenericBox` is used
        instead.

        """

        # Determine the current value of the name_box
        entry_name = get_box_value(name_box)

        # Determine at what index the provided name_box currently is in grid
        index = self.entries_grid.indexOf(name_box)

        # Check if the name_box is valid
        valid = self.is_valid(name_box)

        # Retrieve which value_box is currently there
        cur_box = self.entries_grid.itemAt(index+1).widget()

        # Obtain the widget class associated with this entry_name
        if valid:
            # If the given name is valid, obtain it from dict or use default
            new_box_class = self.entry_types.get(entry_name, GW.LongGenericBox)
        else:
            # If invalid, use a label
            new_box_class = GW.QLabel

        # Check if a modified signal must be emitted later
        emit_signal = False
        if valid or type(cur_box) is not GW.QLabel:
            emit_signal = True

        # If the current value box is not this box type already, replace it
        if type(cur_box) is not new_box_class:
            # If not, create new value box
            new_box = new_box_class()
            new_box.setSizePolicy(QW.QSizePolicy.MinimumExpanding,
                                  QW.QSizePolicy.Fixed)

            # Connect to modified signal
            if valid:
                get_modified_signal(new_box).connect(self.modified)

            # Replace cur_box with new_box
            item = self.entries_grid.replaceWidget(cur_box, new_box)
            cur_box = new_box

            # Close the widget in this item and delete it
            item.widget().close()
            del item

        # If the box is invalid, set its value
        if not valid:
            # If it has a name, set a warning as the value
            if entry_name:
                set_box_value(cur_box, f"Entry <b>{entry_name}</b> is either "
                              "invalid or already in use!")
            # Else, set an empty string as the value
            else:
                set_box_value(cur_box, "")

        # If this entry is valid, perform a few extra tasks
        if valid:
            # Modify tooltip
            cur_box.setToolTip(f"Set value for entry <b>{entry_name}</b>")

            # If this entry has a default value, use it
            if entry_name in self.entry_defaults:
                set_box_value(cur_box, self.entry_defaults[entry_name])

        # Emit modified signal if required
        if emit_signal:
            self.modified.emit()

    # This function updates the entry types dict with a given dict
    def addEntryTypes(self, entry_types):
        """
        Adds the given `entry_types` dict to the dict of entry names that use a
        specific callable when entered.

        Parameters
        ----------
        entry_types : dict
            Dict with the entry names that use a specific callable for
            obtaining the :obj:`~PyQt5.QtWidgets.QWidget` object to be used in
            the entries box.
            The types found in :obj:`~guipy.widgets.type_box_dict` can be given
            instead of a callable as well.
            Optionally, the default value of the widget can
            also be provided.
            This dict is formatted as `{'<entry_name>':
            (<callable>[, <default_value>])}`.

        """

        # Make sure that '' is not in entry_types
        entry_types.pop('', None)

        # Obtain the set of types in entry_types that are currently banned
        banned_types = self.banned_names.intersection(entry_types.keys())

        # Remove these banned entries from entry_types
        for banned_type in banned_types:
            entry_types.pop(banned_type)

        # Split entry_types up into types and defaults
        types = {}
        defaults = {}
        for key, value in entry_types.items():
            # Check if the value is an iterable
            try:
                value = tuple(value)
            except TypeError:
                # If it is not, there is no default value
                pass
            else:
                # If it is, a default value was provided
                defaults[key] = value[1]
                value = value[0]

            # Set proper callable
            types[key] = GW.type_box_dict.get(value, value)

        # Update the current dict of entry types and entry defaults
        self.entry_types.update(types)
        self.entry_defaults.update(defaults)

    # This function adds a dict of default values for given entries
    def addDefaults(self, entry_defaults):
        """
        Adds the given `entry_defaults` dict to the dict of entry names that
        have a default value when entered.

        Note that entry names that do not have a specific value box defined,
        always use a :class:`~guipy.widgets.LongGenericBox`.

        Parameters
        ----------
        entry_defaults : dict
            Dict with the entry names that have a default value when used in
            the entries box.
            This dict is formatted as `{'<entry_name>': <default_value>}`.

        """

        # Make sure that '' is not in entry_defaults
        entry_defaults.pop('', None)

        # Add provided dict to current dict of defaults
        self.entry_defaults.update(entry_defaults)

        # This function adds a list of names to the banned names set
    def addBannedNames(self, banned_names):
        """
        Adds the given `banned_names` list to the set of names that are banned
        from being used as entry names.

        Note that banned names are only used when :meth:`~get_entry_name_box`
        returns an editable combobox.

        Parameters
        ----------
        banned_names : list of str
            List containing entry names that are not allowed.

        """

        # Convert banned_names to a set
        banned_names = sset(banned_names)

        # Make sure that '' is not in banned_names
        banned_names.discard('')

        # Update the current set of banned names
        self.banned_names.update(banned_names)

        # Determine if there are any types in entry_types that are now banned
        banned_types = banned_names.intersection(self.entry_types.keys())

        # Remove all new banned types from entry_types
        for banned_type in banned_types:
            self.entry_types.pop(banned_type)

    # This function retrieves the values of the entries in this entries box
    def get_box_value(self, *value_sig):
        """
        Returns the current values of this entries box as a dict.

        Returns
        -------
        entries_dict : dict
            A dict with the entries currently in this entries box, formatted as
            `{'<entry_name>': <entry_value>}`.

        """

        # Create an empty dict to hold the entry values in
        entries_dict = dict()

        # Loop over all entries in the entries grid and save them to the dict
        for i in range(4, 4+3*self.entryCount(), 3):
            # Obtain the name of this entry
            name_box = self.entries_grid.itemAt(i).widget()
            entry_name = get_box_value(name_box)

            # If the entry_name is invalid, skip this entry
            if not self.is_valid(name_box):
                continue

            # Obtain the value of this entry
            value_box = self.entries_grid.itemAt(i+1).widget()
            entry_value = get_box_value(value_box)

            # Add this entry to the dict
            entries_dict[entry_name] = entry_value

        # Return entries_dict
        return(entries_dict)

    # This function sets the values of the entries in this entries box
    def set_box_value(self, entries_dict, *value_sig):
        """
        Sets the values of the entries in this entries box to the provided
        `entries_dict`.

        Parameters
        ----------
        entries_dict : dict
            A dict containing all entries that must be set in this entries box,
            formatted as `{<entry_name>: <entry_value>}`.

        """

        # Hide the entries box to allow for its values to be set properly
        self.hide()

        # Remove all banned names from entries_dict
        entries_dict = sdict({key: value for key, value in entries_dict.items()
                              if key not in self.banned_names})

        # Create empty dict for all current entries
        cur_entries_dict = {}

        # Remove all entries from the entries box
        for _ in range(self.entryCount()):
            # Obtain the name of the next entry
            name_box = self.entries_grid.itemAt(4).widget()
            entry_name = get_box_value(name_box)

            # Delete this entry if it is not in entries_dict or it is invalid
            if entry_name not in entries_dict or not self.is_valid(name_box):
                self.remove_entry(name_box)
                continue

            # Add this entry to cur_entries_dict
            cur_entries_dict[entry_name] =\
                [self.entries_grid.takeAt(3) for _ in range(3)]

        # Add all entries in entries_dict
        for row, (entry_name, entry_value) in enumerate(entries_dict.items(),
                                                        1):
            # Check if this entry_name is in cur_entries_dict
            if entry_name in cur_entries_dict:
                # If so, put it back into the entries box
                for col, item in enumerate(cur_entries_dict.pop(entry_name)):
                    self.entries_grid.addItem(item, row, col)
            else:
                # If not, add a new empty entry to the entries box
                self.add_entry()

                # Set the name of this entry
                name_box = self.entries_grid.itemAtPosition(row, 1).widget()
                set_box_value(name_box, entry_name)

            # Set the value of this entry
            value_box = self.entries_grid.itemAtPosition(row, 2).widget()
            set_box_value(value_box, entry_value)

        # Show the entries box again now that its values have been set
        self.show()


# Make class for creating editable entry lists
class EditableEntriesBox(EntriesBox):
    """
    Defines the :class:`~EditableEntriesBox` class.

    This widget allows for a series of editable 'list' entries to be maintained
    within a single box, and is basically a Qt-version of a Python dict.
    Entries that must use a specific widget can be added using the
    :meth:`~addEntryTypes` method, whereas disallowed entry names can be added
    with :meth:`~addBannedNames`.

    """

    # This function returns the proper combobox to be used for name entries
    def get_entry_name_box(self):
        # Create name_box
        name_box = GW.EditableComboBox()

        # Initialize and apply special validator
        validator = EntryNameBoxValidator(name_box, self, True)
        name_box.setValidator(validator)

        # Return name_box
        return(name_box)



# -*- coding: utf-8 -*-

"""
Figure Plugin
=============

"""


# %% IMPORTS
# Built-in imports

# Package imports
from qtpy import QtCore as QC, QtWidgets as QW

# GuiPy imports
from guipy import layouts as GL, plugins as GP, widgets as GW
from guipy.plugins.figure.config import MPLConfigPage
from guipy.plugins.figure.widgets import FigureWidget
from guipy.widgets import set_box_value

# All declaration
__all__ = ['Figure']


# %% CLASS DEFINITIONS
# Define class for the Figure plugin
class Figure(GP.BasePluginWidget):
    # Properties
    TITLE = "Figure"
    CONFIG_PAGES = [*GP.BasePluginWidget.CONFIG_PAGES, MPLConfigPage]
    LOCATION = QC.Qt.RightDockWidgetArea
    REQ_PLUGINS = [*GP.BasePluginWidget.REQ_PLUGINS, "Data table"]

    # Initialize Figure plugin
    def __init__(self, *args, **kwargs):
        # Call super constructor
        super().__init__(*args, **kwargs)

        # Extract data_table_obj
        self.data_table = self.req_plugins['Data table']

        # Set up the figure plugin
        self.init()

    # This function sets up the figure plugin
    def init(self):
        # Create a layout
        layout = GL.QVBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)

        # Create a tab widget
        tab_widget = GW.QTabWidget(browse_tabs=True)
        tab_widget.setTabBar(GW.EditableTabBar())
        tab_widget.setMovable(True)
        tab_widget.setTabsClosable(True)

        # Connect tab widget signals
        tab_widget.tabCloseRequested.connect(self.close_tab)

        # Add tab widget to layout
        self.tab_widget = tab_widget
        layout.addWidget(self.tab_widget)

        # Add all actions to the proper menus and toolbars
        self.add_actions()

        # Add a tab to the plugin
        self.add_tab()

    # Override closeEvent to do automatic clean-up
    def closeEvent(self, *args, **kwargs):
        # Block all signals emitted by the tab widget while removing tabs
        self.tab_widget.blockSignals(True)

        # Close all tabs
        for index in reversed(range(self.tab_widget.count())):
            self.close_tab(index)

        # Call super event
        super().closeEvent(*args, **kwargs)

    # This function adds all associated actions to the menus and toolbars
    def add_actions(self):
        # Initialize empty action lists for this plugin
        self.MENU_ACTIONS = {
            **GP.BasePluginWidget.MENU_ACTIONS,
            'File/New': []}

        # Add new figure action to file/new menu
        # TODO: Should shortcut be CTRL+F? Might clash with expectations
        new_tab_act = GW.QAction(
            self, '&Figure',
            shortcut=QC.Qt.CTRL + QC.Qt.Key_F,
            tooltip="New figure",
            triggered=self.add_tab,
            role=GW.QAction.ApplicationSpecificRole)
        self.MENU_ACTIONS['File/New'].append(new_tab_act)

    # This function adds a new figure widget
    @QC.Slot()
    def add_tab(self, name=None):
        # Create a new FigureWidget
        figure = FigureWidget(self.data_table, self.tab_widget.count(), self)

        # If name is None, set it to default
        if name is None:
            name = "figure_%i" % (self.tab_widget.count())

        # Add figure to the tab widget
        index = self.tab_widget.addTab(figure, name)

        # Switch focus to the new tab
        set_box_value(self.tab_widget, index)

    # This function closes a figure widget
    @QC.Slot(int)
    def close_tab(self, index):
        # Obtain the FigureWidget object associated with this index
        figure = self.tab_widget.widget(index)

        # Close this figure
        figure.close()

        # Remove this figure from the tab widget
        self.tab_widget.removeTab(index)



# -*- coding: utf-8 -*-

"""
Histogram Property
==================

"""


# %% IMPORTS
# Package imports
from qtpy import QtCore as QC

# GuiPy imports
from guipy import layouts as GL, widgets as GW
from guipy.plugins.figure.widgets.types.props import BasePlotProp
from guipy.plugins.figure.widgets.types.props.data import (
    Data1DProp, MultiDataNDProp)
from guipy.widgets import get_box_value, get_modified_signal, set_box_value

# All declaration
__all__ = ['HistogramProp', 'MultiHistDataProp']


# %% CLASS DEFINITIONS
# Define 'Histogram' plot property
class HistogramProp(BasePlotProp):
    """
    Provides the definition of the :class:`~HistogramProp` plot property.

    This property contains boxes for setting the number of bins, the value
    range, and the histogram bar colors.

    """

    # Class attributes
    NAME = "Histogram"
    DISPLAY_NAME = "Histogram"
    WIDGET_NAMES = [*BasePlotProp.WIDGET_NAMES, 'n_bins_box',
                    'hist_orient_box', 'hist_cumul_box']

    # This function creates and returns a bins box
    def n_bins_box(self):
        """
        Creates a widget box for setting the number of bins in a histogram and
        returns it.

        """

        # Make spinbox for bin count
        n_bins_box = GW.QSpinBox()
        n_bins_box.setToolTip("Number of bins used for this histogram. Set to "
                              "'auto' to automatically determine this number")
        n_bins_box.setRange(0, 100)
        n_bins_box.setSpecialValueText('auto')

        # Set initial value to 'auto'
        set_box_value(n_bins_box, 0)

        # Return name and box
        return('# of bins', n_bins_box)

    # This function creates a histogram cumulative box
    def hist_cumul_box(self):
        """
        Creates a widget box for setting the cumulative property of this
        histogram and returns it.

        """

        # Make a checkbox
        hist_cumul_box = GW.QCheckBox('Cumulative')
        hist_cumul_box.setToolTip("Toggle the use of a cumulative histogram")

        # Return name and box
        return(hist_cumul_box,)

    # This function creates a histogram orientation box
    def hist_orient_box(self):
        """
        Creates a widget box for setting the orientation property of this
        histogram and returns it.

        """

        # Make a multi radiobutton
        hist_orient_box = GW.MultiRadioButton(['Horizontal', 'Vertical'])
        hist_orient_box.setToolTip("The orientation of the histogram")
        set_box_value(hist_orient_box, 'Vertical')

        # Return name and box
        return('Orientation', hist_orient_box)


# Define 'HistData' plot property
class HistDataProp(Data1DProp):
    """
    Provides the definition of the :class:`~HistDataProp` plot property.

    This property contains boxes for setting the label; X-axis data and color
    for an individual histogram.

    """

    # Class attributes
    NAME = "HistData"
    WIDGET_NAMES = [*Data1DProp.WIDGET_NAMES, 'hist_color_box']

    # This function creates a histogram color box
    def hist_color_box(self):
        """
        Creates a widget box for setting the color of this histogram and
        returns it.

        """

        # Make a color box
        hist_color_box = GW.ColorBox()
        hist_color_box.setToolTip("Color to be used for this histogram")

        # Connect 'applying' signal
        self.options.applying.connect(hist_color_box.set_default_color)

        # Return name and box
        return('Color', hist_color_box)


# Define custom class for setting the data used in the histogram
class MultiHistDataProp(MultiDataNDProp):
    """
    Provides the definition of the :class:`~MultiHistDataProp` plot property.

    This property contains a tab widget with multiple :class`~HistDataProp`
    properties.

    """

    # Class attributes
    NAME = "MultiHistData"
    REQUIREMENTS = [*MultiDataNDProp.REQUIREMENTS, *HistDataProp.REQUIREMENTS,
                    'hist_tab_added']

    # Initialize hist data property
    def __init__(self, *args, **kwargs):
        # Call super constructor
        super().__init__(HistDataProp, *args, **kwargs)

        # Connect signals
        self.tab_widget.tabWasInserted.connect(self.hist_tab_added)


# Define custom class for setting the value range of a histogram
class HistRangeBox(GW.BaseBox):
    # Signals
    rangeChanged = QC.Signal(float, float)

    # Initialize instance of HistRangeBox
    def __init__(self, parent=None, *args, **kwargs):
        # Call super constructor
        super().__init__(parent)

        # Set up histogram range box
        self.init()

    # This function sets up the histogram range box
    def init(self):
        # Make layout for setting the value range of the histogram
        hist_range_layout = GL.QHBoxLayout(self)
        hist_range_layout.setContentsMargins(0, 0, 0, 0)

        # Make a box for setting the value range of the histogram
        # TODO: Maybe use dual lineedits instead to eliminate range problem?
        hist_range_box = GW.DualSpinBox((float, float),
                                        r"<html>&le; X &le;</html>")
        x_min_box, x_max_box = hist_range_box[:]
        x_min_box.setRange(-9999999, 9999999)
        x_min_box.setToolTip("Minimum value to be included in the histogram")
        x_max_box.setRange(-9999999, 9999999)
        x_max_box.setToolTip("Maximum value to be included in the histogram")
        set_box_value(hist_range_box, (0, 0))
        hist_range_box.setEnabled(False)
        self.range_box = hist_range_box

        # Make a checkbox for enabling/disabling the use of this range
        hist_range_flag = GW.QCheckBox()
        hist_range_flag.setToolTip("Toggle the use of a manual histogram value"
                                   " range")
        set_box_value(hist_range_flag, False)
        self.range_flag = hist_range_flag

        # Connect signals for hist_range_flag
        get_modified_signal(hist_range_flag).connect(
            hist_range_box.setEnabled)

        # Add everything together
        hist_range_layout.addWidget(hist_range_flag)
        hist_range_layout.addWidget(hist_range_box)

    # This function retrieves a value of this special box
    def get_box_value(self, *args, **kwargs):
        return(get_box_value(self.range_flag), *get_box_value(self.range_box))

    # This function sets the value of this special box
    def set_box_value(self, value, *args, **kwargs):
        set_box_value(self.range_box, value, *args, **kwargs)



# -*- coding: utf-8 -*-

"""
GenericBoxes
============

"""


# %% IMPORTS
# Built-in imports

# Package imports
from qtpy import QtCore as QC, QtWidgets as QW

# GuiPy imports
from guipy import layouts as GL, widgets as GW
from guipy.widgets import get_box_value, get_modified_signal, set_box_value

# All declaration
__all__ = ['GenericBox', 'LongGenericBox']


# %% CLASS DEFINITIONS
# Define class for creating generic value boxes
class GenericBox(GW.BaseBox):
    """
    Defines the :class:`~GenericBox` class.

    This class is used for making a generic value box that only accepts single
    values, unlike :class:`~LongGenericBox`.
    It currently supports inputs of type bool; float; int; and str.

    """

    # Signals
    modified = QC.Signal([], [object])

    # Initialize the GenericBox class
    def __init__(self, parent=None):
        """
        Initialize an instance of the :class:`~GenericBox` class.

        """

        # Call super constructor
        super().__init__(parent)

        # Create the generic box
        self.init()

    # This property returns the default 'modified' signal
    @property
    def default_modified_signal(self):
        return(self.modified[object])

    # This property returns the types that this box supports
    @property
    def supported_types(self):
        return([bool, float, int, str])

    # This function creates the generic box
    def init(self):
        """
        Sets up the generic value box after it has been initialized.

        """

        # Create the box_layout
        box_layout = GL.QHBoxLayout(self)
        box_layout.setContentsMargins(0, 0, 0, 0)
        self.box_layout = box_layout

        # Create a combobox for the type
        type_box = GW.QComboBox()
        type_box.addItems(sorted([x.__name__ for x in self.supported_types]))
        type_box.setSizePolicy(QW.QSizePolicy.Fixed, QW.QSizePolicy.Fixed)
        set_box_value(type_box, -1)
        get_modified_signal(type_box).connect(self.create_value_box)
        box_layout.addWidget(type_box)
        self.type_box = type_box

        # Create a default value box
        value_box = GW.QWidget()
        box_layout.addWidget(value_box)
        self.value_box = value_box

    # This function is automatically called whenever 'modified' is emitted
    @QC.Slot()
    def modified_signal_slot(self):
        # Emit modified signal
        self.modified[object].emit(get_box_value(self.value_box))

    # This function creates a value_box depending on the type that was selected
    @QC.Slot(str)
    def create_value_box(self, value_type):
        """
        Creates a value box for the provided type `value_type` and replaces the
        current value box with it.

        Parameters
        ----------
        value_type : {'bool'; 'float'; 'int'; 'str'}
            The string of the type of value box that is requested.

        """

        # Create a widget box for the specified value_type
        value_box = GW.type_box_dict[eval(value_type)]()
        value_box.setSizePolicy(QW.QSizePolicy.Expanding, QW.QSizePolicy.Fixed)

        # Set this value_box in the layout
        cur_item = self.box_layout.replaceWidget(self.value_box, value_box)
        cur_item.widget().close()
        del cur_item

        # Save new value_box
        self.value_box = value_box

    # This function retrieves a value of this special box
    def get_box_value(self, *value_sig):
        """
        Returns the current value of the value box.

        Returns
        -------
        value : bool, float, int or str
            The current value of this generic value box.

        """

        # If value_box is currently a QWidget, return None
        if type(self.value_box) is GW.QWidget:
            return(None)
        # Else, return its actual value
        else:
            return(get_box_value(self.value_box, *value_sig))

    # This function sets the value of this special box
    def set_box_value(self, value, *value_sig):
        """
        Sets the value type to `type(value)` and the value to `value`.

        Parameters
        ----------
        value : bool, float, int or str
            The value to use for this generic value box. The type of `value`
            determines which value box must be used.

        """

        set_box_value(self.type_box, type(value).__name__)
        set_box_value(self.value_box, value, *value_sig)


# Define class for creating generic value boxes that supports all types
class LongGenericBox(GenericBox):
    """
    Defines the :class:`~LongGenericBox` class.

    This class is used for making a generic value box that also accepts
    iterables, unlike :class:`~GenericBox`.
    It currently supports inputs of all types found in :obj:`~type_box_dict`.

    """

    # This property returns the types that this box supports
    @property
    def supported_types(self):
        return(GW.type_box_dict.keys())



# -*- coding: utf-8 -*-

"""
Figure Widget
=============

"""


# %% IMPORTS
# Built-in imports

# Package imports
import matplotlib.pyplot as plt
from qtpy import QtCore as QC, QtWidgets as QW

# GuiPy imports
from guipy import layouts as GL, widgets as GW
from guipy.plugins.figure.widgets.canvas import FigureCanvas
from guipy.plugins.figure.widgets.manager import FigureManager
from guipy.plugins.figure.widgets.toolbar import FigureToolbar

# All declaration
__all__ = ['FigureWidget']


# %% CLASS DEFINITIONS
# Define class for the Figure widget
class FigureWidget(GW.QWidget):
    # Initialize FigureWidget
    def __init__(self, data_table_plugin_obj, num, parent=None, *args,
                 **kwargs):
        # Save provided data table object
        self.data_table_plugin = data_table_plugin_obj
        self.num = num

        # Call super constructor
        super().__init__(parent)

        # Set up the figure widget
        self.init(*args, **kwargs)

    # This function sets up the figure widget
    def init(self):
        # Create a layout
        layout = GL.QVBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)

        # Initialize figure
        self.figure, self.canvas, self.manager, self.toolbar =\
            self.create_figure()

        # Add figure toolbar to layout
        layout.addWidget(self.toolbar)

        # Add a separator
        layout.addSeparator()

        # Add figure canvas to layout
        layout.addWidget(self.canvas)

    # This function creates and returns a figure
    def create_figure(self):
        # Create a figure instance
        figure = plt.Figure()

        # Create a canvas for this figure
        canvas = FigureCanvas(figure)

        # Create a manager for this canvas
        manager = FigureManager(canvas, self.num)

        # Create a toolbar for this figure
        toolbar = FigureToolbar(self.data_table_plugin, canvas, self)

        # Return all
        return(figure, canvas, manager, toolbar)



# -*- coding: utf-8 -*-

"""
Widget Utilities
================
Utility functions to make using certain widgets easier.

"""


# %% IMPORTS
# Built-in imports
import os
from sys import platform

# Package imports
import e13tools as e13
from sortedcontainers import SortedSet as sset

# GuiPy imports
from guipy import widgets as GW
from guipy.config import FILE_EXTS, FILE_FORMATS

# All declaration
__all__ = ['getOpenFileName', 'getOpenFileNames', 'getSaveFileName']


# %% DOCSTRINGS
optional_doc =\
    """Optional
    --------
    parent : :obj:`~PyQt5.QtWidgets.QWidget` object or None. Default: None
        The parent to use for creating the file dialog, or *None* for no
        parent.
    caption : str. Default: ''
        The caption (window title) to use for the file dialog.
    basedir : str or None. Default: None
        The base directory the file dialog should start in.
        If *None*, the current directory is used instead.
    filters : list of str or None. Default: None
        List of string extensions to use as the file filters.
        If *None*, no filters are used.
    initial_filter : str. Default: ''
        The string extension that must be used as the initial filter.
        If `filters` is empty, no initial filter will be applied.
        If `filters` does not contain `initial_filter` or if empty, the initial
        filter is set to 'All Supported Files'.
    options : :obj:`~PyQt5.QtWidgets.QFileDialog.Options` object or None. \
        Default: None
        The options to use for this file dialog, or *None* for no options."""


# %% HIDDEN DEFINITIONS
# This function processes all arguments provided to a getXXXFileName function
def _processFileDialogArguments(parent=None, caption='', basedir=None,
                                filters=None, initial_filter='', options=None):
    # If basedir is None, set it to the current directory
    if basedir is None:
        basedir = os.getcwd()

    # Check what filters are given and act accordingly
    if filters is None:
        # If no filters are given, set filters and initial_filter to ''
        filters = ''
        initial_filter = ''
    else:
        # Loop over all filters and grab their formats and exts
        formats = sset({FILE_FORMATS.get(file_filter.lower(), '')
                       for file_filter in filters})
        exts = sset({FILE_EXTS.get(file_filter.lower(), '')
                    for file_filter in filters})

        # Remove all empty strings from formats and exts
        if '' in formats:
            formats.remove('')
        if '' in exts:
            exts.remove('')

        # Combine all exts together to a single string
        exts = ' '.join(exts)

        # Convert formats to a list
        formats = list(formats)

        # Add 'All Supported Files' to formats
        formats.append("All supported files (%s)" % (exts))

        # Add 'All Files' to formats
        formats.append("All files (*)")

        # Combine all formats into a single string
        filters = ';;'.join(formats)

        # Process given initial_filter
        if "*"+initial_filter.lower() in exts.split():
            # If initial_filter is in the filters, set it to the proper format
            initial_filter = FILE_FORMATS[initial_filter.lower()]
        else:
            # Else, set to 'All Supported Files'
            initial_filter = formats[-2]

    # Do not use native dialog on Linux, as it is pretty bad
    if platform.startswith('linux'):
        if options is None:
            options = GW.QFileDialog.DontUseNativeDialog
        else:
            options = options | GW.QFileDialog.DontUseNativeDialog

    # Create dict with all arguments
    args_dict = {
        'parent': parent,
        'caption': caption,
        'directory': basedir,
        'filter': filters,
        'initialFilter': initial_filter}

    # If options is not None, add it as well
    if options is not None:
        args_dict['options'] = options

    # Return args_dict
    return(args_dict)


# %% FUNCTION DEFINITIONS
# Define custom getOpenFileName function that automatically applies filters
@e13.docstring_substitute(optional=optional_doc)
def getOpenFileName(*args, **kwargs):
    """
    Wrapper for the :func:`~PyQt5.QtWidgets.QFileDialog.getOpenFileName`
    function, which automatically applies several standard options and makes
    using filters easier.

    %(optional)s

    Returns
    -------
    filename : str
        The filename that was chosen by the user.
    selected_filter : str
        The filter that was used by the user.

    """

    # Process the input arguments
    args_dict = _processFileDialogArguments(*args, **kwargs)

    # Open the file opening system and return the result
    return(GW.QFileDialog.getOpenFileName(**args_dict))


# Define custom getOpenFileNames function that automatically applies filters
@e13.docstring_substitute(optional=optional_doc)
def getOpenFileNames(*args, **kwargs):
    """
    Wrapper for the :func:`~PyQt5.QtWidgets.QFileDialog.getOpenFileNames`
    function, which automatically applies several standard options and makes
    using filters easier.

    %(optional)s

    Returns
    -------
    filenames : list of str
        The filenames that were chosen by the user.
    selected_filter : str
        The filter that was used by the user.

    """

    # Process the input arguments
    args_dict = _processFileDialogArguments(*args, **kwargs)

    # Open the file opening system and return the result
    return(GW.QFileDialog.getOpenFileNames(**args_dict))


# Define custom getSaveFileName function that automatically applies filters
@e13.docstring_substitute(optional=optional_doc)
def getSaveFileName(*args, **kwargs):
    """
    Wrapper for the :func:`~PyQt5.QtWidgets.QFileDialog.getSaveFileName`
    function, which automatically applies several standard options and makes
    using filters easier.

    %(optional)s

    Returns
    -------
    filename : str
        The filename that was chosen by the user.
    selected_filter : str
        The filter that was used by the user.

    """

    # Process the input arguments
    args_dict = _processFileDialogArguments(*args, **kwargs)

    # Open the file saving system and return the result
    return(GW.QFileDialog.getSaveFileName(**args_dict))



# -*- coding: utf-8 -*-

"""
Base Layouts
============
Provides a collection of custom :class:`~PyQt5.QtWidgets.QLayout` base classes
that allow for certain layouts to be standardized.

"""


# %% IMPORTS
# Package imports
from qtpy import QtWidgets as QW

# All declaration
__all__ = ['QFormLayout', 'QGridLayout', 'QHBoxLayout', 'QVBoxLayout']


# %% CLASS DEFINITIONS
# Make subclass of QW.QFormLayout to add separator function
class QFormLayout(QW.QFormLayout):
    # This function adds a horizontal separator to the layout
    def addSeparator(self):
        # Create frame object
        frame = QW.QFrame()

        # Set the frame shape
        frame.setFrameShape(frame.HLine)
        frame.setFrameShadow(frame.Sunken)

        # Add frame to this layout
        self.addRow(frame)


# Make subclass of QW.QGridLayout to add separator function
class QGridLayout(QW.QGridLayout):
    # This function adds a separator to the layout
    def addSeparator(self, *, row=None, column=None):
        # Create frame object
        frame = QW.QFrame()

        # Set the frame shape
        frame.setFrameShape(frame.VLine if row is None else frame.HLine)
        frame.setFrameShadow(frame.Sunken)

        # Add frame to this layout
        if row is None:
            self.addWidget(frame, 0, column, -1, 1)
        else:
            self.addWidget(frame, row, 0, 1, -1)


# Make subclass of QW.QHBoxLayout to add separator function
class QHBoxLayout(QW.QHBoxLayout):
    # This function adds a vertical separator to the layout
    def addSeparator(self):
        # Create frame object
        frame = QW.QFrame()

        # Set the frame shape
        frame.setFrameShape(frame.VLine)
        frame.setFrameShadow(frame.Sunken)

        # Add frame to this layout
        self.addWidget(frame)


# Make subclass of QW.QVBoxLayout to add separator function
class QVBoxLayout(QW.QVBoxLayout):
    # This function adds a horizontal separator to the layout
    def addSeparator(self):
        # Create frame object
        frame = QW.QFrame()

        # Set the frame shape
        frame.setFrameShape(frame.HLine)
        frame.setFrameShadow(frame.Sunken)

        # Add frame to this layout
        self.addWidget(frame)



# -*- coding: utf-8 -*-

"""
General Config
==============

"""


# %% IMPORTS
# Built-in imports
from ast import literal_eval

# Package imports
from qtpy import QtCore as QC, QtGui as QG, QtWidgets as QW
from sortedcontainers import SortedDict as sdict

# GuiPy imports
from guipy import layouts as GL, widgets as GW
from guipy.config import BaseConfigPage
from guipy.widgets import get_box_value, set_box_value

# All declaration
__all__ = ['GeneralConfigPage']


# %% HELPER DEFINITIONS
# Define special ComboBox for setting country and language values
class LocaleComboBox(GW.BaseBox):
    # Signals
    modified = QC.Signal([], [int])

    # Override constructor
    def __init__(self, locale_type, parent=None):
        # Handle provided locale_type
        if(locale_type == 'country'):
            self.f_ToString = QC.QLocale.countryToString
        else:
            self.f_ToString = QC.QLocale.languageToString

        # Call super constructor
        super().__init__(parent)

        # Create the combobox
        self.init()

    # This property returns the default 'modified' signal
    @property
    def default_modified_signal(self):
        return(self.modified[int])

    # This function creates the combobox
    def init(self):
        # Create layout
        layout = GL.QHBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)

        # Create combobox
        combobox = GW.QComboBox()
        layout.addWidget(combobox)
        self.combobox = combobox

    # This function is automatically called whenever 'modified' is emitted
    @QC.Slot()
    def modified_signal_slot(self):
        # Emit current value
        self.modified[int].emit(get_box_value(self))

    # Override addItems to store additional item information
    def addItems(self, items):
        # Loop over all items and add them to the combobox
        for i, item in enumerate(items):
            self.combobox.addItem(self.f_ToString(item))
            self.combobox.setItemData(i, item)

    # This function retrieves a value of this combobox
    def get_box_value(self, *value_sig):
        # Obtain the current index of this combobox
        index = get_box_value(self.combobox, int)

        # Obtain the corresponding user data
        value = self.combobox.itemData(index)

        # Return value
        return(value)

    # This function sets the value of this combobox
    def set_box_value(self, value, *value_sig):
        # Convert provided value to appropriate string
        string = self.f_ToString(value)

        # Set value of this combobox
        set_box_value(self.combobox, string)


# %% CLASS DEFINITIONS
# Define config page for setting the general config
class GeneralConfigPage(BaseConfigPage):
    # Define class attributes
    NAME = 'General'

    # This function sets up the General config page
    def init(self):
        # Create layout
        layout = GL.QVBoxLayout(self)

        # Create 'General' group box
        general_group = GW.QGroupBox("General")
        layout.addWidget(general_group)
        general_layout = GL.QFormLayout(general_group)

        # Add combobox for language-settings
        language_box = LocaleComboBox('language')
        language_box.addItems(list({QC.QLocale.system().language(),
                                    QC.QLocale.English}))
        language_box.setToolTip("Set the language used for representing "
                                "strings and values.")
        self.add_config_entry('language', language_box, True)
        general_layout.addRow("Language", language_box)

    # This function parses and processes a config section, and returns it
    def decode_config(self, section_dict):
        # Initialize empty dict of parsed config values
        config_dict = sdict()

        # Decode all values in section_dict
        for key, value in section_dict.items():
            # Add all values to config dict using literal_eval
            config_dict[key] = literal_eval(value)

        # Return config_dict
        return(config_dict)

    # This function returns a dict containing the default config values
    def get_default_config(self):
        # Create default dict
        default_dict = sdict({
            'language': QC.QLocale.system().language()})

        # Return default_dict
        return(default_dict)

    # This function returns its config section, as required by config parser
    def encode_config(self, config_dict):
        # Initialize empty dict of section config values
        section_dict = sdict()

        # Loop over all arguments in config and encode them in
        for key, value in config_dict.items():
            section_dict[key] = '{!r}'.format(value)

        # Return section_dict
        return(section_dict)

    # This function applies the currently stored config
    def apply_config(self, config_dict):
        # Create new QLocale
        locale = QC.QLocale(config_dict['language'])

        # Set this as the new default
        QC.QLocale.setDefault(locale)



# -*- coding utf-8 -*-

"""
Figure
======

"""


# %% IMPORTS
# Import base modules
from . import plugin
from .plugin import *

# Import subpackages
from . import widgets

# All declaration
__all__ = ['plugin', 'widgets']
__all__.extend(plugin.__all__)

# Author declaration
__author__ = "Ellert van der Velden (@1313e)"



# -*- coding: utf-8 -*-

"""
ItemsBoxes
==========

"""


# %% IMPORTS
# Built-in imports


# Package imports
from qtpy import QtCore as QC, QtGui as QG, QtWidgets as QW

# GuiPy imports
from guipy import INT_TYPES
from guipy import layouts as GL, widgets as GW
from guipy.widgets import get_box_value, get_modified_signal, set_box_value

# All declaration
__all__ = ['GenericItemsBox', 'ItemsBox']


# %% CLASS DEFINITIONS
# Make class for creating item lists
class GenericItemsBox(GW.BaseBox):
    """
    Defines the :class:`~GenericItemsBox` class.

    This widget allows for a series of generic 'list' items to be maintained
    within a single box, and is basically a Qt-version of a Python list.
    Unlike :class:`~ItemsBox`, this widget allows for an unlimited number of
    boxes to be created.

    """

    # Signals
    modified = QC.Signal([], [list])

    # Initialize the ItemsBox class
    def __init__(self, item_type=None, parent=None):
        """
        Initialize an instance of the :class:`~GenericItemsBox` class.

        Optional
        --------
        item_type : callable or None. Default: None
            A callable that returns the :obj:`~PyQt5.QtWidgets.QWidget` object
            that must be used for all items in this items box.
            The types found in :obj:`~guipy.widgets.type_box_dict` can be given
            instead of a callable as well.
            If *None*, the type is unspecified.
        parent : :obj:`~PyQt5.QtWidgets.QWidget` object or None. Default: None
            The parent widget to use for this items box or *None* for no
            parent.

        """

        # Call super constructor
        super().__init__(parent, auto_connect=False)

        # Create the items box
        self.init(item_type)

    # This property returns the default 'modified' signal
    @property
    def default_modified_signal(self):
        return(self.modified[list])

    # This property returns the number of items in the box
    def itemCount(self):
        return(self.items_layout.count())

    # This function creates the items box
    def init(self, item_type):
        """
        Sets up the items box after it has been initialized.

        """

        # Obtain the item_box
        if item_type is None:
            self.item_box = GW.LongGenericBox
        else:
            self.item_box = GW.type_box_dict.get(item_type, item_type)

        # Set the height of a single item
        self.entry_height = 24

        # Create the box_layout
        box_layout = GL.QVBoxLayout(self)
        box_layout.setContentsMargins(0, 0, 0, 0)

        # Create the items_layout
        items_layout = GL.QVBoxLayout()
        items_layout.setContentsMargins(0, 0, 0, 0)
        box_layout.addLayout(items_layout)
        self.items_layout = items_layout

        # Add an 'Add'-button
        add_but = GW.QToolButton()
        add_but.setFixedSize(self.entry_height, self.entry_height)
        add_but.setToolTip("Add new item")
        get_modified_signal(add_but).connect(self.add_item)
        box_layout.addWidget(add_but)

        # If this theme has an 'add' icon, use it
        if QG.QIcon.hasThemeIcon('add'):
            add_but.setIcon(QG.QIcon.fromTheme('add'))
        # Else, use a simple plus
        else:
            add_but.setText('+')

        # Set size policy
        self.setSizePolicy(QW.QSizePolicy.Preferred, QW.QSizePolicy.Fixed)

    # This function is automatically called whenever 'modified' is emitted
    @QC.Slot()
    def modified_signal_slot(self):
        self.modified[list].emit(GenericItemsBox.get_box_value(self))

    # This function is called whenever a new item is added
    @QC.Slot()
    def add_item(self):
        """
        Adds a new item to the items box.

        """

        # Create an item layout
        item_layout = GL.QHBoxLayout()
        item_layout.setContentsMargins(0, 0, 0, 0)

        # Create a generic box for this item
        item_box = self.item_box()
        item_box.setToolTip("Set value of this item")
        get_modified_signal(item_box).connect(self.modified)

        # Create a 'Delete'-button
        del_but = GW.QToolButton()
        del_but.setFixedSize(self.entry_height, self.entry_height)
        del_but.setToolTip("Delete this item")
        get_modified_signal(del_but).connect(
            lambda: self.remove_item(item_layout))

        # If this theme has a 'remove' icon, use it
        if QG.QIcon.hasThemeIcon('remove'):
            del_but.setIcon(QG.QIcon.fromTheme('remove'))
        # Else, use a standard icon
        else:
            del_but.setIcon(del_but.style().standardIcon(
                QW.QStyle.SP_DialogCloseButton))

        # Add a new row to the items layout
        item_layout.addWidget(del_but)
        item_layout.addWidget(item_box)
        self.items_layout.addLayout(item_layout)

        # Emit signal if item_box is not LongGenericBox
        if not isinstance(item_box, GW.LongGenericBox):
            self.modified.emit()

    # This function is called whenever an item must be removed
    @QC.Slot(GW.QComboBox)
    def remove_item(self, item_layout):
        """
        Removes the item associated with the provided `item_layout` from the
        items box.

        """

        # Obtain the value in the layout
        item_box = item_layout.itemAt(1).widget()
        item_value = get_box_value(item_box)

        # If item_value is not None, a modified signal must be emitted later
        emit_signal = (item_value is not None)

        # Remove corresponding item layout
        self.items_layout.removeItem(item_layout)

        # Remove the two items in item_layout
        for _ in range(2):
            # Remove the current item at index 0
            item = item_layout.takeAt(0)

            # Close the widget in this item and delete it
            item.widget().close()
            del item

        # Emit modified signal if required
        if emit_signal:
            self.modified.emit()

    # This function retrieves the values of the items in this items box
    def get_box_value(self, *value_sig):
        """
        Returns the current values of this items box as a list.

        Returns
        -------
        value_list : list
            A list with the values of all items currently in this items box.

        """

        # Create an empty list to hold the item values in
        value_list = []

        # Loop over all items in the items form and save them to the list
        for i in range(self.itemCount()):
            # Obtain the value of this entry
            item_layout = self.items_layout.itemAt(i)
            item_box = item_layout.itemAt(1).widget()
            item_value = get_box_value(item_box)

            # Add this item to the list if it is not None
            if item_value is not None:
                value_list.append(item_value)

        # Return value_list
        return(value_list)

    # This function sets the values of the items in this items box
    def set_box_value(self, value_list, *value_sig):
        """
        Sets the values of the items in this items box to the provided
        `value_list`.

        Parameters
        ----------
        value_list : list
            A list containing the values of all items that must be set in this
            items box.

        """

        # Hide the items box to allow for its values to be set properly
        self.hide()

        # Remove all items from the items box, registering their values
        cur_items_dict = {}
        for _ in range(self.itemCount()):
            # Remove this item and obtain its value
            layout = self.items_layout.takeAt(0)
            value = get_box_value(layout.itemAt(1).widget())

            # Check if it is required later
            if value in value_list:
                # If so, store for later
                cur_items_dict[value] = layout
            else:
                # If not, delete it
                self.remove_item(layout)

        # Add all items in value_list
        for row, value in enumerate(value_list):
            # Check if this value is in cur_items_dict
            if value in cur_items_dict:
                # If so, put it back into the items box
                self.items_layout.insertLayout(row, cur_items_dict.pop(value))
            else:
                # If not, add a new item
                self.add_item()

                # Set the value of this item
                item_layout = self.items_layout.itemAt(row)
                item_box = item_layout.itemAt(1).widget()
                set_box_value(item_box, value)

        # Show the items box again now that its values have been set
        self.show()


# Make class for creating item lists
class ItemsBox(GW.BaseBox):
    """
    Defines the :class:`~ItemsBox` class.

    This widget allows for a series of 'list' items to be maintained within a
    single box, and is basically a Qt-version of a Python list.
    Unlike :class:`~GenericItemsBox`, this widget is provided with a
    pre-determined set of boxes that it will use.

    """

    # Signals
    modified = QC.Signal([], [list])

    # Initialize the ItemsBox class
    def __init__(self, item_types, layout='horizontal', parent=None):
        """
        Initialize an instance of the :class:`~ItemsBox` class.

        Parameters
        ----------
        item_types : list of callables
            A list of callables that each return the
            :obj:`~PyQt5.QtWidgets.QWidget` object that must be used in this
            items box.
            The types found in :obj:`~guipy.widgets.type_box_dict` can be given
            instead of a callable as well.

        Optional
        --------
        layout : {'horizontal'; 'vertical'}. Default: 'horizontal'
            The layout that this items box must have.
            If 'horizontal' all items will be aligned on a single row.
            If 'vertical', all items will be aligned on a single column.
        parent : :obj:`~PyQt5.QtWidgets.QWidget` object or None. Default: None
            The parent widget to use for this items box or *None* for no
            parent.

        """

        # Call super constructor
        super().__init__(parent)

        # Create the items box
        self.init(item_types, layout)

    # Override __getitem__ to return the requested item
    def __getitem__(self, key):
        # If key is an integer, return the corresponding item
        if isinstance(key, INT_TYPES):
            # Try to return the requested item
            try:
                return(self.items[key])
            # If that cannot be done, raise IndexError
            except IndexError:
                raise IndexError("Index out of range")

        # If key is a slice object, return everything that is requested
        elif isinstance(key, slice):
            return(*map(self.__getitem__, range(*key.indices(self.N))),)

        # Else, raise TypeError
        else:
            raise TypeError("Index must be of type 'int' or 'slice', "
                            "not type %r" % (type(key).__name__))

    # This property returns the default 'modified' signal
    @property
    def default_modified_signal(self):
        return(self.modified[list])

    # This property returns the number of items in the box
    def itemCount(self):
        return(self.N)

    # This function creates the items box
    def init(self, item_types, layout):
        """
        Sets up the items box after it has been initialized.

        """

        # Set the height of a single entry
        self.entry_height = 24

        # Create empty list of items
        self.items = []

        # Obtain the proper layout
        if layout in ('h', 'horizontal', 'r', 'row'):
            layout = GL.QHBoxLayout(self)
        elif layout in ('v', 'vertical', 'c', 'col', 'column'):
            layout = GL.QVBoxLayout(self)
        else:
            raise ValueError
        layout.setContentsMargins(0, 0, 0, 0)

        # Initialize all items
        for item_type in item_types:
            # Create item
            item = GW.type_box_dict.get(item_type, item_type)()

            # Add item to items list and layout
            self.items.append(item)
            layout.addWidget(item)

        # Save the number of items
        self.N = len(self.items)

    # This function is automatically called whenever 'modified' is emitted
    @QC.Slot()
    def modified_signal_slot(self):
        self.modified[list].emit(ItemsBox.get_box_value(self))

    # This function retrieves the values of the items in this items box
    def get_box_value(self, *value_sig):
        """
        Returns the current values of this items box as a list.

        Returns
        -------
        items_list : object or list
            If `value_sig` contains a single index, the value of the
            corresponding item.
            Otherwise, a list with the values of all items in this items box.

        """

        # Check if value_sig contains an integer
        if value_sig and isinstance(value_sig[0], INT_TYPES):
            # If so, return the value of that specific item
            return(get_box_value(self.items[value_sig[0]], *value_sig[1:]))
        else:
            # If not, return the values of all items
            return(list(map(get_box_value, self.items)))

    # This function sets the values of the items in this items box
    def set_box_value(self, value_list, *value_sig):
        """
        Sets the values of the items in this items box to the provided
        `items_list`.

        Parameters
        ----------
        value_list : object of list
            If `value_sig` contains a single index, the value that must be set
            for corresponding item.
            Otherwise, a list containing the values for all items that must be
            set in this items box.

        """

        # Check if value_sig contains an integer
        if value_sig and isinstance(value_sig[0], INT_TYPES):
            # If so, set the value of that specific item
            set_box_value(self.items[value_sig[0]], value_list, *value_sig[1:])
        else:
            # If not, set the values of all items
            for item, value in zip(self.items, value_list):
                set_box_value(item, value)



# -*- coding: utf-8 -*-

"""
Config Manager
==============

"""


# %% IMPORTS
# Built-in imports
from configparser import ConfigParser
import os
from os import path

# Package imports
from qtpy import QtCore as QC
from sortedcontainers import SortedDict as sdict

# GuiPy imports
from guipy import APP_NAME, CONFIG_DIR, CONFIG_NAME, widgets as GW
from guipy.config.dialog import ConfigDialog
from guipy.widgets import get_box_value, set_box_value

# All declaration
__all__ = ['CONFIG']


# %% CLASS DEFINITIONS
# Define config manager
class ConfigManager(object):
    # Initialize config manager
    def __init__(self):
        # Initialize different configuration dicts and parsers
        self.config = sdict()
        self.parser = ConfigParser(interpolation=None)
        self.parser.optionxform = str
        self.config_pages = sdict()

    # Initialize config manager for actual use in GuiPy
    def _init(self, parent=None):
        # Save provided parent
        self.parent = parent

        # Initialize config dialog
        self.config_dialog = ConfigDialog(self, parent=parent)

        # Connect signals
        self.config_dialog.applying.connect(self.apply_config)
        self.config_dialog.discarding.connect(self.discard_config)
        self.config_dialog.resetting.connect(self.reset_config)

        # Read in the configuration file
        self.read_config()

        # Add core config pages
        self._add_config_pages()

        # Make sure that the locale of the parent and dialog are set properly
        self.parent.setLocale(QC.QLocale())

    # This function returns the value of a specific config
    def get_option(self, section, option):
        return(self.config[section][option])

    # This function retrieves the config folder
    def _get_config_dir(self):
        """
        Determines the directory that contains the configuration files of
        *GuiPy*, located in the user's home directory, and returns it.
        If it does not exist yet, it will be created first.

        Returns
        -------
        config_dir : str
            Path to the configuration directory of *GuiPy*.

        """

        # Determine the directory that should contain the GuiPy config
        config_dir = path.join(path.expanduser('~'), CONFIG_DIR)

        # If this directory does not exist, make it
        if not path.exists(config_dir):
            os.mkdir(config_dir)

        # Return config_dir
        return(config_dir)

    # This function retrieves the config file
    def _get_config_file(self):
        """
        Determines the main configuration file of *GuiPy*, located in the
        configuration directory, and returns it.

        Returns
        -------
        config_file : str
            Path to the main configuration file of *GuiPy*.

        """

        # Obtain the configuration directory
        config_dir = self._get_config_dir()

        # Determine the file that should contain the GuiPy main config
        config_file = path.join(config_dir, CONFIG_NAME)

        # If the config file does not exist, make an empty one
        if not path.exists(config_file):
            open(config_file, 'w').close()
            os.chmod(config_file, 0o644)

        # Return config_file
        return(config_file)

    # This function adds the core config pages to the config manager
    def _add_config_pages(self):
        """
        Adds the core config pages to the config manager, which are required by
        *GuiPy* itself.

        """

        # Import all required config pages
        from guipy.config import GeneralConfigPage

        # Add 'General' config page
        GeneralConfigPage()

    # This function adds a config page to the config manager
    def add_config_page(self, config_page):
        """
        Adds a provided :obj:`~guipy.config.BaseConfigPage` `config_page` to
        the config manager.

        """

        # Add config page to dialog
        self.config_dialog.add_config_page(config_page)

        # Obtain the section name of this config page
        section_name = config_page.section_name

        # Add the config_page to the config_pages dict
        self.config_pages[section_name] = config_page

        # Check if the section already exists, add it if not
        if not self.parser.has_section(section_name):
            self.parser.add_section(section_name)

        # Obtain the associated config section from the parser
        config_section = self.parser[section_name]

        # Parse the config section belonging to this config page
        parsed_dict = config_page.decode_config(config_section)

        # Obtain the default config of this section
        config_dict = config_page.get_default_config()

        # Update default dict with parsed dict
        config_dict.update(parsed_dict)

        # Add this config_dict to the global config
        self.config[section_name] = config_dict

        # Set this config_dict as the current values
        config_page.apply_config(config_dict)
        set_box_value(config_page, config_dict)

        # Revert flags that were set due to the config page being modified
        self.config_dialog.disable_apply_button()
        config_page.restart_flag = False

        # Get updated config section from the config page
        config_section = config_page.encode_config(config_dict)

        # Update the parser with the config_section
        self.parser[section_name] = config_section

    # This function reads in the configuration file from the config folder
    def read_config(self, config_file=None):
        """
        Reads the main configuration file of *GuiPy* given by `config_file`
        into this config manager.

        Optional
        --------
        config_file : str or None. Default: None
            The path to the configuration file to read. If *None*, the default
            is used.

        """

        # Retrieve the configuration file of GuiPy if required
        if config_file is None:
            config_file = self._get_config_file()

        # Read configuration
        self.parser.read(config_file)

    # This function write the current config to the configuration file
    def write_config(self, config_file=None):
        """
        Writes all configuration values in this config manager to the main
        configuration file of *GuiPy* given by `config_file`.

        Optional
        --------
        config_file : str or None. Default: None
            The path to the configuration file to write. If *None*, the default
            is used.

        """

        # Retrieve the configuration file if required
        if config_file is None:
            config_file = self._get_config_file()

        # Write current parser to this file
        with open(config_file, 'w') as config_file:
            self.parser.write(config_file)

    # This function applies the current config
    def apply_config(self):
        # Initialize restart_flag
        restart_flag = False

        # Loop over all config pages and obtain their values
        for name, page in self.config_pages.items():
            # Obtain config_dict
            config_dict = get_box_value(page)
            page.apply_config(config_dict)

            # Store config_dict in saved config and parser
            self.config[name] = config_dict
            self.parser[name] = page.encode_config(config_dict)

            # Add restart_flag of page to current restart_flag
            restart_flag += page.restart_flag
            page.restart_flag = False

        # Save config to file
        self.write_config()

        # If restart_flag is True, show warning dialog
        if restart_flag:
            self.show_restart_warning()

    # This function discards the current config and resets it the saved values
    def discard_config(self):
        # Loop over all config pages and discard their values
        for name, page in self.config_pages.items():
            # Obtain config_dict
            config_dict = self.config[name]

            # Set value of page to this dict
            set_box_value(page, config_dict)

    # This function resets the config to its default values
    def reset_config(self):
        # Initialize restart_flag
        restart_flag = False

        # Loop over all config pages and reset their values
        for name, page in self.config_pages.items():
            # Obtain default config_dict
            config_dict = page.get_default_config()

            # Set value of page to this dict
            page.apply_config(config_dict)
            set_box_value(page, config_dict)

            # Store config_dict in saved config and parser
            self.config[name] = config_dict
            self.parser[name] = page.encode_config(config_dict)

            # Add restart_flag of page to current restart_flag
            restart_flag += page.restart_flag
            page.restart_flag = False

        # Save config to file
        self.write_config()

        # If restart_flag is True, show warning dialog
        if restart_flag:
            self.show_restart_warning()

    # This function displays a warning dialog that GuiPy must be restarted
    def show_restart_warning(self):
        GW.QMessageBox.warning(
            None, "Restart warning",
            (f"One or more configuration settings have been modified that "
             f"require {APP_NAME} to be restarted to take effect. Please "
             f"restart {APP_NAME} now."))


# %% IMPORT SCRIPT
# Initialize config manager as CONFIG
CONFIG = ConfigManager()



# -*- coding: utf-8 -*-

"""
NPZ Formatter
=============

"""


# %% IMPORTS
# Built-in imports
from ast import literal_eval

# Package imports
import numpy as np
import pandas as pd

# GuiPy imports
from guipy.plugins.data_table.formatters import BaseFormatter

# All declaration
__all__ = ['NPZFormatter']


# %% CLASS DEFINITIONS
# Define Formatter for .npz-files
class NPZFormatter(BaseFormatter):
    # Class attributes
    TYPE = "NumPy Binary Archive"
    EXTS = ['.npz']

    # Define the export to npz function
    def exporter(self, data_table, filepath):
        # Obtain the data in the data table
        data = data_table.model._data

        # Make a dictionary that contains the data of all columns
        data_dict = {"(%i, %r)" % (i, name): column.values
                     for i, (name, column) in enumerate(data.items())}

        # Save data table as NumPy Binary Archive
        np.savez(filepath, **data_dict)

    # Define the import from npz function
    def importer(self, filepath, parent=None):
        # Load the archive
        archive = np.load(filepath)

        # Convert the loaded archive to a dict
        data_dict = dict(archive)

        # Close the archive
        archive.close()

        # Create empty column dict and naming dict
        column_dict = {}
        name_dict = {}

        # Loop over all data arrays in data_dict and obtain their indices
        for i, (key, data) in enumerate(data_dict.items()):
            # Try to convert key to index and name
            try:
                index, name = literal_eval(key)

            # If that does not work, use default values
            except ValueError:
                index = i
                name = key

            # Add index, name and data to the proper dicts
            column_dict[index] = data
            name_dict[index] = name

        # Create a data frame
        data_frame = pd.DataFrame(column_dict)

        # Sort the columns on their indices
        data_frame.sort_index(1, inplace=True)

        # Rename the columns to their proper names
        data_frame.rename(columns=name_dict, inplace=True)

        # Return data_frame
        return(data_frame)



# -*- coding: utf-8 -*-

"""
Widgets Core
============
Provides a collection of utility functions and the :class:`~BaseBox` class
definition, which are core to the functioning of all widgets.

"""


# %% IMPORTS
# Built-in import
from inspect import isclass

# Package imports
from qtpy import QtCore as QC, QtGui as QG, QtWidgets as QW

# GuiPy imports
from guipy import FLOAT_TYPES, INT_TYPES, STR_TYPES
from guipy.widgets.base import (
    QLabel as GW_QLabel, QTabWidget as GW_QTabWidget, QWidget as GW_QWidget)

# All declaration
__all__ = ['BaseBox', 'DualBaseBox', 'get_box_value', 'get_modified_signal',
           'set_box_value']


# %% CLASS DEFINITIONS
# Make base class for custom boxes
# As QW.QWidget is a strict class (in C++), this cannot be an ABC
class BaseBox(GW_QWidget):
    """
    Defines the :class:`~BaseBox` base class.

    This class is used by many custom :class:`~PyQt5.QtWidgets.QWidget` classes
    as their base. It defines the :attr:`~modified` signal, which is
    automatically connected to any widget that changes its state, unless
    `auto_connect` is set set *False*.

    """

    # Define modified signal
    modified = QC.Signal()

    # Initialize the BaseBox class
    def __init__(self, *args, auto_connect=True, **kwargs):
        # Call super constructor
        super().__init__(*args, **kwargs)

        # Store auto_connect
        self.auto_connect = auto_connect

        # If the 'modified_signal_slot' slot is available, connect it
        if hasattr(self, 'modified_signal_slot'):
            self.modified.connect(self.modified_signal_slot)

    # Override childEvent to connect signals if child has a modified signal
    def childEvent(self, event):
        """
        Special :meth:`~PyQt5.QtCore.QObject.childEvent` event that
        automatically connects the default modified signal of any widget that
        becomes a child of this widget.

        """

        # If this event involved a child being added, check child object
        if self.auto_connect and (event.type() == QC.QEvent.ChildAdded):
            # Obtain child object
            child = event.child()

            # Try to obtain the modified signal of this child
            try:
                signal = get_modified_signal(child)
            # If this fails, it does not have one
            except NotImplementedError:
                pass
            # If this succeeds, connect it to the 'modified' signal
            else:
                signal.connect(self.modified[()])

        # Call and return super method
        return(super().childEvent(event))

    # This function connects a given box to the modified signal
    def connect_box(self, box):
        """
        Connect the default modified signal of the provided `box` to this
        widget's :attr:`~modified` signal.

        """

        # Check if the given box is a child of this box and skip if so
        if box in self.children():
            return

        # Obtain the modified signal of the given box
        signal = get_modified_signal(box)

        # Connect the signals
        signal.connect(self.modified[()])

    # Define get_box_value method
    def get_box_value(self, *value_sig):
        """
        Obtain the value of this widget and return it.

        """

        raise NotImplementedError(self.__class__)

    # Define set_box_value method
    def set_box_value(self, value, *value_sig):
        """
        Set the value of this widget to `value`.

        """

        raise NotImplementedError(self.__class__)


# Define DualBaseBox class for making dual widgets
class DualBaseBox(BaseBox):
    # Override __getitem__ to return the left and/or right widget
    def __getitem__(self, key):
        # If key is an integer, return the corresponding widget
        if isinstance(key, INT_TYPES):
            # If key is 0 or -2, return left_box
            if key in (0, -2):
                return(self.left_box)
            # Else, if key is 1 or -1, return right_box
            elif key in (1, -1):
                return(self.right_box)
            # Else, raise IndexError
            else:
                raise IndexError("Index out of range")

        # If key is a slice object, return everything that is requested
        elif isinstance(key, slice):
            return(*map(self.__getitem__, range(*key.indices(2))),)

        # Else, raise TypeError
        else:
            raise TypeError("Index must be of type 'int' or 'slice', not type "
                            "%r" % (type(key).__name__))

    # This function retrieves a value of this special box
    def get_box_value(self, *value_sig):
        """
        Returns the current values of this dual widget as a tuple.

        Returns
        -------
        value : tuple
            A tuple containing the values of the widgets, formatted as
            `(left, right)`.

        """

        # Get values of both widgets
        if(len(value_sig) == 2):
            # If value_sig contains exactly 2 elements, use them separately
            left_sig, right_sig = value_sig

            # Make sure that left_sig and right_sig are iterables to unpack
            if not isinstance(left_sig, (list, tuple, set)):
                left_sig = (left_sig,)
            if not isinstance(right_sig, (list, tuple, set)):
                right_sig = (right_sig,)

            # Get values
            return(get_box_value(self.left_box, *left_sig),
                   get_box_value(self.right_box, *right_sig))
        else:
            # Else, use value_sig for both
            return(get_box_value(self.left_box, *value_sig),
                   get_box_value(self.right_box, *value_sig))

    # This function sets the value of this special box
    def set_box_value(self, value, *value_sig):
        """
        Sets the current value of the dual widget to `value`.

        Parameters
        ----------
        value : tuple
            A tuple containing the values of the widgets, formatted as
            `(left, right)`.

        """

        # Set values of both widgets
        if(len(value_sig) == 2):
            # If value_sig contains exactly 2 elements, use them separately
            left_sig, right_sig = value_sig

            # Make sure that left_sig and right_sig are iterables to unpack
            if not isinstance(left_sig, (list, tuple, set)):
                left_sig = (left_sig,)
            if not isinstance(right_sig, (list, tuple, set)):
                right_sig = (right_sig,)

            # Set values
            set_box_value(self.left_box, *left_sig)
            set_box_value(self.right_box, *right_sig)
        else:
            # Else, use value_sig for both
            set_box_value(self.left_box, value[0], *value_sig)
            set_box_value(self.right_box, value[1], *value_sig)

    # Override closeEvent to make sure both widgets are deleted when closed
    def closeEvent(self, event):
        # Close both widgets
        self.left_box.close()
        self.right_box.close()

        # Call super method
        return(super().closeEvent(event))


# %% FUNCTION DEFINITIONS
# This function gets the value of a provided box
def get_box_value(box, *value_sig, no_custom=False):
    """
    Retrieves the value of the provided widget `box` and returns it.

    If `box` has the `get_box_value()` method defined (always the case for
    instances of :class:`~BaseBox`), it will be used instead. If this raises a
    :class:`~NotImplementedError`, the method is skipped.
    If `no_custom` is *True*, the `get_box_value()` method is also skipped.
    Note that this only works if `box` is not an instance of :class:`~BaseBox`.

    Parameters
    ----------
    box : :obj:`~PyQt5.QtWidgets.QWidget` object
        The widget whose value must be returned.
    value_sig : positional arguments of object
        The signature of the value of `box` that must be returned.
        If empty or invalid, the default value is returned.
        If `box` has the `get_box_value()` method defined, this argument is
        passed to it.

    Optional
    --------
    no_custom : bool. Default: False
        Whether to skip any present `get_box_value()` methods of the provided
        `box`. Note that setting this to *True* for instances of
        :class:`~BaseBox` will result in an error.

    Returns
    -------
    box_value : obj
        The value of the requested `box`.

    """

    # Custom boxes (get_box_value()-method)
    if hasattr(box, 'get_box_value') and not no_custom:
        # Try to use the custom get_box_value()-method
        try:
            return(box.get_box_value(*value_sig))
        # If that fails, proceed to use a normal function
        except NotImplementedError:
            pass

    # Values (QAbstractSpinBox)
    if isinstance(box, QW.QAbstractSpinBox):
        return(box.value())

    # Actions, Bools/Buttons (QAction, QAbstractButton)
    elif isinstance(box, (QW.QAction, QW.QAbstractButton)):
        if box.isCheckable() and str not in value_sig:
            return(box.isChecked())
        else:
            return(box.text())

    # Items (QComboBox)
    elif isinstance(box, QW.QComboBox):
        return(box.currentIndex() if int in value_sig else box.currentText())

    # Tabs (QTabWidget)
    elif isinstance(box, QW.QTabWidget):
        if int in value_sig:
            return(box.currentIndex())
        elif str in value_sig:
            return(box.tabText(box.currentIndex()))
        else:
            return(box.currentWidget())

    # Strings (QLineEdit)
    elif isinstance(box, QW.QLineEdit):
        return(box.text())

    # Labels (QLabel)
    elif isinstance(box, QW.QLabel):
        for attr in ['movie', 'picture', 'pixmap']:
            if getattr(box, attr)() is not None:
                return(getattr(box, attr)())
        else:
            return(box.text())

    # If none applies, raise error
    else:
        raise NotImplementedError("Custom boxes must provide their own "
                                  "'get_box_value()'-method! (%s)"
                                  % (box.__class__))


# This function gets the emitted signal when a provided box is modified
def get_modified_signal(box, *signal_sig):
    """
    Retrieves a signal of the provided widget `box` that indicates that `box`
    has been modified and returns it.

    If `box` has the `default_modified_signal` attribute defined, it will be
    returned if `signal_sig` is empty.
    If `box` has the `modified` signal defined (always the case for instances
    of :class:`~BaseBox`), it will be returned instead.

    Parameters
    ----------
    box : :obj:`~PyQt5.QtWidgets.QWidget` object
        The widget whose modified signal must be retrieved.
    signal_sig : positional arguments of object
        The signature of the modified signal that is requested.
        If empty or invalid, the default modified signal is returned.

    Returns
    -------
    modified_signal : :obj:`~PyQt5.QtCore.pyqtBoundSignal` object
        The requested modified signal of `box`.

    """

    # Custom boxes (default_modified_signal attribute)
    if hasattr(box, 'default_modified_signal') and not signal_sig:
        return(box.default_modified_signal)

    # Custom boxes (modified signal)
    elif hasattr(box, 'modified'):
        return(box.modified.__getitem__(signal_sig))

    # Values (QAbstractSpinBox)
    elif isinstance(box, QW.QAbstractSpinBox):
        return(box.valueChanged)

    # Actions (QAction)
    elif isinstance(box, QW.QAction):
        return(box.toggled if box.isCheckable() else box.triggered)

    # Bools/Buttons (QAbstractButton)
    elif isinstance(box, QW.QAbstractButton):
        return(box.toggled if box.isCheckable() else box.clicked)

    # Items (QComboBox)
    elif isinstance(box, QW.QComboBox):
        return(box.currentIndexChanged if int in signal_sig else
               box.currentTextChanged)

    # Strings (QLineEdit)
    elif isinstance(box, QW.QLineEdit):
        return(box.textChanged)

    # Labels (QLabel)
    elif isinstance(box, GW_QLabel):
        return(box.contentsChanged)
    elif isinstance(box, QW.QLabel):
        raise NotImplementedError("Default QW.QLabel has no modified signal "
                                  "defined. Use QW_QLabel instead!")

    # If none applies, raise error
    else:
        raise NotImplementedError("Custom boxes must provide their own "
                                  "'modified' signal! (%s)" % (box.__class__))


# This function sets the value of a provided box
def set_box_value(box, value, *value_sig, no_custom=False):
    """
    Sets the value of the provided widget `box` to `value`.

    If `box` has the `set_box_value()` method defined (always the case for
    instances of :class:`~BaseBox`), it will be used instead. If this raises a
    :class:`~NotImplementedError`, the method is skipped.
    If `no_custom` is *True*, the `set_box_value()` method is also skipped.
    Note that this only works if `box` is not an instance of :class:`~BaseBox`

    Parameters
    ----------
    box : :obj:`~PyQt5.QtWidgets.QWidget` object
        The widget whose value must be set.
    value : obj
        The value that must be set in the provided `box`.
        If `box` has the `set_box_value()` method defined, this argument is
        passed it.
    value_sig : positional arguments of object
        The signature of the value that must be set in the provided `box`.
        This argument is only used if the signature cannot be derived from
        `value`.
        If `box` has the `set_box_value()` method defined, this argument is
        passed to it.

    Optional
    --------
    no_custom : bool. Default: False
        Whether to skip any present `set_box_value()` methods of the provided
        `box`. Note that setting this to *True* for instances of
        :class:`~BaseBox` will result in an error.

    """

    # Custom boxes (set_box_value()-method)
    if hasattr(box, 'set_box_value') and not no_custom:
        # Try to use the custom set_box_value()-method
        try:
            box.set_box_value(value, *value_sig)
        # If that fails, proceed to use a normal function
        except NotImplementedError:
            pass
        # If that succeeds, return
        else:
            return

    # Values (QAbstractSpinBox)
    if isinstance(box, QW.QAbstractSpinBox):
        box.setValue(value)

    # Actions, Bools/Buttons (QAction, QAbstractButton)
    elif isinstance(box, (QW.QAction, QW.QAbstractButton)):
        if isinstance(value, STR_TYPES):
            box.setText(value)
        else:
            box.setChecked(value)

    # Items (QComboBox)
    elif isinstance(box, QW.QComboBox):
        if isinstance(value, INT_TYPES):
            box.setCurrentIndex(value)
        else:
            index = box.findText(value)
            if(index != -1):
                box.setCurrentIndex(index)
            elif box.isEditable():
                # Obtain current index and text of this combobox
                index = box.currentIndex()
                text = box.currentText()

                # Block all signals coming from this combobox
                blocked = box.blockSignals(True)

                # Set the current index to -1 and set the text to value
                box.setCurrentIndex(-1)
                box.setCurrentText(value)

                # No longer block signals (unless they were blocked before)
                box.blockSignals(blocked)

                # Emit the proper signals if their values actually changed
                if(text != value):
                    box.currentTextChanged.emit(value)
                if(index != -1):
                    box.currentIndexChanged.emit(-1)

    # Tabs (QTabWidget)
    elif isinstance(box, QW.QTabWidget):
        if isinstance(value, INT_TYPES):
            box.setCurrentIndex(value)
        elif isinstance(value, STR_TYPES) and isinstance(box, GW_QTabWidget):
            index = box.tabNames.index(value)
            box.setCurrentIndex(index)
        else:
            box.setCurrentWidget(value)

    # Strings (QLineEdit)
    elif isinstance(box, QW.QLineEdit):
        box.setText(value)

    # Labels (QLabel)
    elif isinstance(box, QW.QLabel):
        if isinstance(value, STR_TYPES):
            box.setText(value)
        elif isinstance(value, FLOAT_TYPES):
            box.setNum(value)
        elif isinstance(value, QG.QMovie):
            box.setMovie(value)
        elif isinstance(value, QG.QPicture):
            box.setPicture(value)
        elif isinstance(value, QG.QPixmap):
            box.setPixmap(value)
        else:
            raise TypeError("QLabel does not support the given type")

    # If none applies, raise error
    else:
        raise NotImplementedError("Custom boxes must provide their own "
                                  "'set_box_value()'-method! (%s)"
                                  % (box.__class__))



# -*- coding: utf-8 -*-

"""
Base Plot Properties
====================

"""


# %% IMPORTS
# GuiPy imports
from guipy import layouts as GL
from guipy.widgets import get_modified_signal

# All declaration
__all__ = ['BasePlotProp']


# %% CLASS DEFINITIONS
# Define BasePlotProp base class
class BasePlotProp(GL.QFormLayout):
    """
    Provides a base class definition that must be subclassed by all figure
    plot properties.

    Every plot property has a `name` and `display_name` associated with it;
    a list of `requirements`; and a list of all `widget_names`.

    The `name` is the name/string used internally to identify this property.

    The `display_name` is the text used in the header of the property group
    box.

    The `requirements` specify all attributes that are required by this
    property, which must be provided by the class initializing it.

    The `widget_names` are the names of the methods that must be called to
    receive the widgets for this plot property.
    These methods return the widget name (as shown in the prop layout) and the
    widget itself.
    The widgets will also be registered as instance attributes with the same
    name as their construction method.

    The `track_values` is a bool that specifies if all widgets in this
    plot property should have their values tracked. This is *True* by default.

    """

    # Define class attributes
    NAME = ""
    DISPLAY_NAME = ""
    REQUIREMENTS = ['options']
    WIDGET_NAMES = []
    TRACK_VALUES = True

    # Initialize this plot property
    def __init__(self, parent=None, **kwargs):
        # Call super constructor
        super().__init__(parent)

        # Set up the plot property
        self.init(**kwargs)

    # This function sets up the plot property
    def init(self, **kwargs):
        # Save all provided kwargs as instance attributes
        for name, value in kwargs.items():
            setattr(self, name, value)

        # Save specific methods of provided options
        self.get_option = self.options.get_option

        # Initialize empty dict of registered widgets
        self.widgets = {}

        # Loop over all items in self.widget_names
        for widget_name in self.WIDGET_NAMES:
            # Obtain the function to get this widget
            func = getattr(self, widget_name)

            # Call the function
            out = func()

            # Register the widget
            widget = out[-1]
            self.widgets[widget_name] = widget

            # Add widget as an options entry if requested
            if self.TRACK_VALUES:
                self.options.add_options_entry(widget)
            # Else, solely connect the modified signal to the apply button
            else:
                get_modified_signal(widget).connect(
                    self.options.enable_apply_button)

            # Add widget to the layout
            self.addRow(*out)

    # Create a close method
    def close(self, *args, **kwargs):
        # Remove all widgets from options entries dict and close them
        for widget in self.widgets.values():
            # If this property uses an option entry, remove the widget
            if self.TRACK_VALUES:
                self.options.remove_options_entry(widget)

            # Close widget
            widget.close()



# -*- coding: utf-8 -*-

"""
Application
===========
Contains the core application function for using *GuiPy*.

"""


# %% IMPORTS
# Import base modules
from . import main_window
from .main_window import *
from .start import main

# All declaration
__all__ = ['main_window', 'main']
__all__.extend(main_window.__all__)

# Author declaration
__author__ = "Ellert van der Velden (@1313e)"



# -*- coding: utf-8 -*-

"""
Figure Toolbar
==============

"""


# %% IMPORTS
# Built-in imports
from qtpy import QtCore as QC, QtWidgets as QW

# Package imports
from matplotlib.backends.backend_qt5agg import NavigationToolbar2QT

# GuiPy imports
from guipy import widgets as GW
from guipy.plugins.figure.widgets.options import FigureOptionsDialog
from guipy.widgets import get_modified_signal, set_box_value

# All declaration
__all__ = ['FigureToolbar']


# %% CLASS DEFINITIONS
# Custom FigureToolbar class
class FigureToolbar(NavigationToolbar2QT, GW.QToolBar):
    # Initialize FigureToolbar class
    def __init__(self, data_table_plugin_obj, canvas, parent=None):
        # Save provided data table plugin object
        self.data_table_plugin = data_table_plugin_obj

        # Call super constructor
        super().__init__(canvas, parent, coordinates=False)

        # Retrieve parent methods
        self.get_parent_methods()

        # Set up figure toolbar
        self.init()

    # This function sets up the figure toolbar
    def init(self):
        # Create pointer in figure manager to this toolbar
        self.canvas.manager.toolbar = self

        # Initialize options dialog
        self.options_dialog = FigureOptionsDialog(self)
        self.labels = ['>>> Figure &options...', '<<< Figure &options...']

        # Obtain the first action in the toolbar
        action_0 = self.actions()[0]

        # Create button for showing/hiding extra options
        dialog_but = GW.QPushButton()
        set_box_value(dialog_but, self.labels[self.options_dialog.isHidden()])
        dialog_but.setToolTip("Toggle the figure options menu")
        get_modified_signal(dialog_but).connect(self.toggle_options_dialog)
        dialog_but.setSizePolicy(QW.QSizePolicy.Fixed, QW.QSizePolicy.Fixed)
        self.insertWidget(action_0, dialog_but)
        self.dialog_but = dialog_but

        # Insert separator
        self.insertSeparator(action_0)

        # Add separator
        self.addSeparator()

        # Add a label that contains the coordinates of the figure
        coord_label = GW.QLabel('')
        self.addWidget(coord_label)
        self.message.connect(coord_label.setText)

    # This function toggles the options dialog
    @QC.Slot()
    def toggle_options_dialog(self):
        # Toggle the options dialog
        self.options_dialog.setVisible(self.options_dialog.isHidden())
        set_box_value(self.dialog_but,
                      self.labels[self.options_dialog.isHidden()])



# -*- coding: utf-8 -*-

"""
Plot Properties Core
====================
Collects all the registered figure plot properties into a single dict.

"""


# %% IMPORTS
# Built-in imports
from importlib import import_module
import os
from os import path

# Package imports
from sortedcontainers import SortedDict as sdict

# All declaration
__all__ = ['PLOT_PROPS', 'register_plot_prop']


# %% GLOBALS
# Define dict of data table formatters
PLOT_PROPS = sdict()


# %% FUNCTION DEFINITIONS
# This function registers a plot property
def register_plot_prop(plot_prop_class):
    """
    Registers a provided plot property class `plot_prop_class` for use in
    *GuiPy*.

    All plot properties must be registered with this function in order to be
    used.

    Parameters
    ----------
    plot_prop_class : :class:`~guipy.plugins.figure.widgets.types.\
        props.BasePlotProp` subclass
        The plot property class to use for formatting a plot property.

    """

    # Register the plot property
    PLOT_PROPS[plot_prop_class.NAME] = plot_prop_class


# This function imports all pre-defined plot properties and registers them
def _import_plot_props():
    """
    Imports and registers all pre-defined plot properties for use in *GuiPy*.

    """

    # Obtain the path to this directory
    dirpath = path.dirname(__file__)

    # Obtain a list of all files in this directory
    filenames = next(os.walk(dirpath))[2]

    # Remove __init__.py, base.py and core.py
    filenames.remove('__init__.py')
    filenames.remove('base.py')
    filenames.remove('core.py')

    # Loop over all files remaining
    for filename in filenames:
        # Obtain full module name
        modname = "%s.%s" % (__package__, filename[:-3])

        # Import this module
        mod = import_module(modname)

        # Register everything in __all__ as a plot property
        for prop in mod.__all__:
            prop_class = getattr(mod, prop)
            register_plot_prop(prop_class)


# Import all defined plot properties
_import_plot_props()



# -*- coding: utf-8 -*-

"""
Config Dialog
=============

"""


# %% IMPORTS
# Built-in imports
import re

# Package importsd
from qtpy import QtCore as QC, QtWidgets as QW
from sortedcontainers import SortedDict as sdict

# GuiPy imports
from guipy import layouts as GL, widgets as GW
from guipy.widgets import get_modified_signal

# All declaration
__all__ = ['ConfigDialog']


# %% CLASS DEFINITIONS
# Define config dialog
class ConfigDialog(GW.QDialog):
    """
    Defines the :class:`~ConfigDialog` class for *GuiPy*.

    This dialog serves as a general configuration skeleton, to which other
    configuration pages can be added through *GuiPy*'s configuration manager.

    """

    # Define class attributes
    NAME = "&Configuration"

    # Create applying, discarding and resetting signals
    applying = QC.Signal()
    discarding = QC.Signal()
    resetting = QC.Signal()

    # Initialize ConfigDialog class
    def __init__(self, config_manager_obj, *args, **kwargs):
        """
        Initialize an instance of the :class:`~ConfigDialog` class.

        Parameters
        ----------
        config_manager_obj : :obj:`~guipy.config.ConfigManager` object
            The config manager object that this config dialog will be connected
            to.

        """

        # Save provided config_manager_obj
        self.config_manager = config_manager_obj

        # Call super constructor
        super().__init__(*args, **kwargs)

        # Set up the config dialog
        self.init()

    # This function shows the config dialog
    @QC.Slot()
    def __call__(self):
        """
        Qt slot that shows the config dialog in the center of the main window.

        """

        # Show it
        self.show()

        # Move the options window to the center of the main window
        self.move(self.parent().geometry().center()-self.rect().center())

    # This function sets up the configuration dialog
    def init(self):
        """
        Sets up the dialog used for setting the configuration in *GuiPy*.

        """

        # Set properties of configuration dialog
        self.setWindowTitle(self.NAME.replace('&', ''))
        self.setWindowFlags(
            QC.Qt.MSWindowsOwnDC |
            QC.Qt.Dialog |
            QC.Qt.WindowTitleHint |
            QC.Qt.WindowSystemMenuHint |
            QC.Qt.WindowCloseButtonHint)

        # Create a window layout
        layout = GL.QVBoxLayout(self)

        # Create a splitter widget for this window
        splitter = GW.QSplitter()
        splitter.setChildrenCollapsible(False)
        layout.addWidget(splitter)

        # Create a contents widget
        contents = GW.QListWidget()
        contents.setMovement(GW.QListView.Static)
        contents.setSpacing(1)
        splitter.addWidget(contents)
        splitter.setStretchFactor(0, 1)
        self.contents = contents

        # Create a sections widget
        sections = GW.QStackedWidget()
        splitter.addWidget(sections)
        splitter.setStretchFactor(1, 4)
        self.sections = sections

        # Connect signals
        contents.currentRowChanged.connect(sections.setCurrentIndex)

        # Create empty dict of config pages
        self.config_pages = sdict()

        # Add a buttonbox
        button_box = QW.QDialogButtonBox()
        button_box.clicked.connect(self.buttonWasPressed)
        layout.addWidget(button_box)
        self.button_box = button_box

        # Add an 'Ok' button
        ok_but = button_box.addButton(button_box.Ok)
        ok_but.setToolTip("Apply current changes and close the configuration "
                          "window")

        # Add a 'Cancel' button
        cancel_but = button_box.addButton(button_box.Cancel)
        cancel_but.setToolTip("Discard current changes and close the "
                              "configuration window")
        self.cancel_but = cancel_but

        # Add an 'Apply' button
        apply_but = button_box.addButton(button_box.Apply)
        apply_but.setToolTip("Apply current configuration changes")
        self.apply_but = apply_but
        self.disable_apply_button()

        # Add a 'Reset' button
        reset_but = button_box.addButton(button_box.Reset)
        reset_but.setToolTip("Reset configuration values to defaults")

        # Create a slot dict for all buttons
        self.slot_dict = {
            button_box.AcceptRole: [
                self.apply_options,
                self.close],
            button_box.RejectRole: [
                self.discard_options,
                self.close],
            button_box.ApplyRole: [
                self.apply_options],
            button_box.ResetRole: [
                self.reset_options]}

        # Set the size of the dialog
        self.resize(800, 600)

    # This function adds a new BasicConfigPage to the dialog
    @QC.Slot(QW.QWidget)
    def add_config_page(self, config_page):
        """
        Adds a provided `config_page` to this dialog, allowing it to be
        modified.
        The name of the config page determines where it will appear in the
        dialog.

        Parameters
        ----------
        config_page : :obj:`~guipy.config.BaseConfigPage` object
            The config page object that must be added to this dialog.

        """

        # Add a stretch to the layout of the page
        config_page.layout().addStretch()

        # Create scrollarea for the config page
        scroll_area = GW.QScrollArea(self)
        scroll_area.setWidgetResizable(True)
        scroll_area.setWidget(config_page)

        # Obtain the page sections of this config page
        main, sub, tab = self.get_page_sections(config_page.section_name)

        # Obtain the full page section for this config page
        page_section = self.config_pages.setdefault(
            main, sdict()).setdefault(sub, sdict())

        # Check if this section existed before
        if(len(page_section) == 0):
            # Add the config page as a widget to the page_section and sections
            self.sections.addWidget(scroll_area)
            self.contents.addItem(main)

            # If this is the first section in contents, select it
            if(self.contents.count() == 1):
                self.contents.setCurrentRow(0)

        elif(len(page_section) == 1):
            # If so, a tab widget will be required
            tab_widget = GW.QTabWidget()
            page_section[''] = tab_widget

            # Obtain the index of the current widget at this section
            prev_page = page_section.values()[1]
            index = self.sections.indexOf(prev_page)

            # Remove this widget
            self.sections.removeWidget(prev_page)

            # Add the tab_widget
            self.sections.insertWidget(index, tab_widget)

            # Add the previous widget and the new config page to the tab_widget
            tab_widget.addTab(prev_page, page_section.keys()[1])
            tab_widget.addTab(scroll_area, tab)

        else:
            # Obtain the current tab widget
            tab_widget = page_section['']
            tab_widget.addTab(scroll_area, tab)

        # Connect signals
        get_modified_signal(config_page).connect(self.enable_apply_button)

        # Add this config page to the proper page section
        page_section[tab] = scroll_area

    # This function determines the config dialog sections a name belongs to
    def get_page_sections(self, section_name):
        """
        Determines which page (sub)sections the given `section_name` would
        belong to and returns it.

        Parameters
        ----------
        section_name : str
            The name of a specific config section.

        Returns
        -------
        main_section : str
            The name of the main section for `section_name`.
        sub_section : str
            The name of the sub section for `section_name`.
        tab_name : str
            The name of the tab on the main section for `section_name`.

        """

        # Create a regex pattern for finding the names
        pattern = r"([^/:]+)(/)?(?(2)([^:]+))(:)?(?(4)(.+))"

        # Match the given section_name
        match = re.match(pattern, section_name)

        # Obtain the different names
        main, sub, tab = match.group(1, 3, 5)

        # Add the sub to main if it is not None
        # TODO: Add the subsection system
        if sub is not None:
            main = f"{main}/{sub}"
            sub = None

        # If tab is None, set it to 'General'
        if tab is None:
            tab = 'General'

        # Return names
        return(main, sub, tab)

    # This function is called whenever a button is pressed
    @QC.Slot(QW.QAbstractButton)
    def buttonWasPressed(self, button):
        """
        Handles the actions that should be carried out when the provided
        `button` is pressed.

        """

        # Obtain the role of the provided button
        but_role = self.button_box.buttonRole(button)

        # Obtain the button's slots that must be called
        slots = self.slot_dict[but_role]

        # Call all slots defined in slots
        for slot in slots:
            slot()

    # This function enables the apply button
    @QC.Slot()
    def enable_apply_button(self):
        """
        Qt slot that enables the apply button at the bottom of the figure
        options dialog.
        The apply button is enabled if at least one change has been made to any
        figure option.

        """

        self.apply_but.setEnabled(True)

    # This function disables the apply button
    @QC.Slot()
    def disable_apply_button(self):
        """
        Qt slot that disables the apply button at the bottom of the figure
        options dialog.
        The apply button is disabled whenever no changes have been made to any
        figure option.

        """

        self.apply_but.setEnabled(False)

    # This function applies the new figure options values
    @QC.Slot()
    def apply_options(self):
        """
        Applies all current changes to all configuration values.

        """

        # Emit the applying signal
        self.applying.emit()

        # Disable the apply button
        self.disable_apply_button()

    # This function discards all changes to the figure options
    @QC.Slot()
    def discard_options(self):
        """
        Discards the current changes to all configuration values and sets them
        back to their saved values.

        """

        # Emit the discarding signal
        self.discarding.emit()

        # Disable the apply button
        self.disable_apply_button()

    # This function discards all changes to the figure options
    @QC.Slot()
    def reset_options(self):
        """
        Resets all configuration values back to their defaults.

        """

        # Emit the resetting signal
        self.resetting.emit()

        # Disable the apply button
        self.disable_apply_button()

    # Override reject to discard all options if called
    def reject(self):
        self.discard_options()
        super().reject()



# -*- coding: utf-8 -*-

"""
Comboboxes
==========

"""


# %% IMPORTS
# Built-in imports

# Package imports
from qtpy import QtCore as QC, QtGui as QG, QtWidgets as QW

# GuiPy imports
from guipy import layouts as GL, widgets as GW

# All declaration
__all__ = ['ComboBoxValidator', 'DualComboBox', 'EditableComboBox',
           'create_combobox']


# %% CLASS DEFINITIONS
# Define the ComboBoxValidator class
class ComboBoxValidator(QG.QRegularExpressionValidator):
    # Initialize the ComboBoxValidator class
    def __init__(self, combobox_obj, regexp=None, parent=None):
        """
        Initialize an instance of the :class:`~ComboBoxValidator` class.

        Parameters
        ----------
        combobox_obj : :obj:`~PyQt5.QtWidgets.ComboBox` object
            Combobox object for which the editable line must be validated.

        Optional
        --------
        regexp : str or None. Default: None
            The regular expression pattern to use for validating an input
            string if it is not found in `combobox_obj`.
            If *None*, no regular expression is used.
        parent : :obj:`~PyQt5.QtCore.QObject` object or None. Default: None
            The parent object to use for this validator or *None* for no
            parent.

        """

        # Save the completer of the provided combobox
        self.completer = combobox_obj.completer()

        # Check provided regexp
        if regexp is None:
            # If regexp is None, set the pattern to one that rejects all
            regexp = r"$.^"

        # Call super constructor
        super().__init__(QC.QRegularExpression(regexp), parent)

    # Override validate to first check the combobox completer
    def validate(self, string, pos):
        # Check if string is already in the completions list
        index = self.completer.completionModel().index(0, 0)
        match = self.completer.completionModel().match(
            index, QC.Qt.EditRole, string, flags=QC.Qt.MatchExactly)

        # Check if there is a match
        if not match:
            # If not, set the completion prefix in the completer
            self.completer.setCompletionPrefix(string)

        # Obtain the current completion string
        completion = self.completer.currentCompletion()

        # Check the completion string against the given one and act accordingly
        if completion:
            # If the completion string is not empty, check if it matches
            if match and string:
                # If so, it is acceptable
                state = self.Acceptable
            else:
                # Else, it is intermediate
                state = self.Intermediate

            # Return the state, string and pos
            return(state, string, pos)
        else:
            # If the completion string is empty, use the regular expression
            return(super().validate(string, pos))


# Make class with two comboboxes
class DualComboBox(GW.DualBaseBox):
    """
    Defines the :class:`~DualComboBox` class.

    """

    # Signals
    modified = QC.Signal([], [int, int], [int, str], [str, int],
                         [str, str])

    # Initialize the DualComboBox class
    def __init__(self, editable=(False, False), sep=None, parent=None):
        """
        Initialize an instance of the :class:`~DualComboBox` class.

        Optional
        --------
        editable : tuple of bool. Default: (False, False)
            A tuple containing the editability of each combobox.
        sep : str or None. Default: None
            The string that must be used as a separator between the two
            comboboxes. If *None*, no separator is used.
        parent : :obj:`~PyQt5.QtWidgets.QWidget` object or None. Default: None
            The parent widget to use for this dual combobox or *None* for no
            parent.

        """

        # Call super constructor
        super().__init__(parent)

        # Create the dual combobox
        self.init(editable, sep)

    # This property returns the default 'modified' signal
    @property
    def default_modified_signal(self):
        return(self.modified[str, str])

    # This function sets up the dual combobox
    def init(self, editable, sep):
        """
        Sets up the dual combobox after it has been initialized.

        """

        # Create the box_layout
        box_layout = GL.QHBoxLayout(self)
        box_layout.setContentsMargins(0, 0, 0, 0)

        # Create two comboboxes with the provided editability
        # LEFT
        left_box = EditableComboBox() if editable[0] else GW.QComboBox()
        box_layout.addWidget(left_box)
        self.left_box = left_box

        # RIGHT
        right_box = EditableComboBox() if editable[1] else GW.QComboBox()
        box_layout.addWidget(right_box)
        self.right_box = right_box

        # If sep is not None, create label and add it
        if sep is not None:
            sep_label = GW.QLabel(sep)
            sep_label.setSizePolicy(QW.QSizePolicy.Fixed, QW.QSizePolicy.Fixed)
            box_layout.insertWidget(1, sep_label)

    # This function is automatically called whenever 'modified' is emitted
    @QC.Slot()
    def modified_signal_slot(self):
        # Emit modified signal with proper types
        for types in [(int, int), (int, str), (str, int), (str, str)]:
            self.modified[types[0], types[1]].emit(
                *DualComboBox.get_box_value(self, *types))


# Create custom QComboBox class that is editable
class EditableComboBox(GW.QComboBox):
    """
    Defines the :class:`~QEditableComboBox` class.

    This class makes the :class:`~guipy.widgets.QComboBox` class editable.

    """

    # Create focusLost signal
    focusLost = QC.Signal([int], [str])

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setEditable(True)
        self.setInsertPolicy(self.NoInsert)
        self.completer().setCompletionMode(QW.QCompleter.PopupCompletion)
        self.completer().setFilterMode(QC.Qt.MatchContains)

    # Override focusOutEvent to emit signal whenever triggered
    def focusOutEvent(self, event):
        # Emit focusLost signal
        self.focusLost[int].emit(self.currentIndex())
        self.focusLost[str].emit(self.currentText())

        # Call super method
        return(super().focusOutEvent(event))


# %% FUNCTION DEFINITIONS
# Create small function factory for comboboxes
def create_combobox(items, editable=False, parent=None):
    """
    Creates and returns a function that, when called, creates the requested
    :obj:`~guipy.widgets.QComboBox` and populates it with the `items` provided.

    Parameters
    ----------
    items : list of str
        List containing the items that the combobox must have when created.

    Optional
    --------
    editable : bool. Default: False
        Whether the combobox that is created must be editable.
    parent : :obj:`~PyQt5.QtWidgets.QWidget` object or None. Default: None
        The parent widget to use for thecombobox or *None* for no parent.

    """

    # Create function
    def func():
        # Obtain the proper combobox
        if editable:
            combobox = EditableComboBox(parent=parent)
        else:
            combobox = GW.QComboBox(parent=parent)

        # Add items
        combobox.addItems(items)

        # Return combobox
        return(combobox)

    # Return func
    return(func)



# -*- coding: utf-8 -*-

"""
Figure Config
=============

"""


# %% IMPORTS
# Built-in imports
from ast import literal_eval
from importlib import import_module

# Package imports
from cycler import cycler
import matplotlib as mpl
from matplotlib import rcParams, rcParamsDefault
from qtpy import QtCore as QC, QtGui as QG, QtWidgets as QW
from sortedcontainers import SortedDict as sdict

# GuiPy imports
from guipy import layouts as GL, plugins as GP, widgets as GW
from guipy.widgets import (
    create_combobox, get_box_value, type_box_dict, set_box_value)

# All declaration
__all__ = ['MPLConfigPage']


# %% HELPER DEFINITIONS
# Create custom togglebox class
class AutoToggleBox(GW.ToggleBox):
    def __init__(self, widget, untoggled):
        # Call super constructor
        super().__init__(widget)

        # Save untoggled
        self.untoggled = untoggled

    # Override get_box_value to return untoggled when the bool is False
    def get_box_value(self, *value_sig):
        # Obtain value of checkbox
        check = get_box_value(self.checkbox)

        # If True, value is obtained from the widget
        if check:
            value = get_box_value(self.widget, *value_sig)
        else:
            value = self.untoggled

        # Return value
        return(value)

    # Override set_box_value to allow for untoggled to toggle the box
    def set_box_value(self, value, *value_sig):
        # Check if value is untoggled
        if(value == self.untoggled):
            # If so, toggle the widget
            set_box_value(self.checkbox, False)
        else:
            # Else, set the checkbox to True and set the value
            set_box_value(self.checkbox, True)
            set_box_value(self.widget, value, *value_sig)


# Create custom combobox class that allows for bools to be used as well
class BoolComboBox(GW.QComboBox):
    def __init__(self, items, *args, **kwargs):
        # Call super constructor
        super().__init__(*args, **kwargs)

        # Initialize uses_bools flag
        self.uses_bools = False

        # Add provided items
        self.addItems(items)

    # Override addItems to handle bools differently
    def addItems(self, items):
        # Make empty lists to separate normal items and bools
        std_items = []
        bool_items = []

        # Loop over all items and separate them into the proper lists
        for item in items:
            # Check if this item is a bool or None
            if item in (True, False, None):
                # If so, save this
                self.uses_bools = True

                # Convert to string and add to bool_items
                bool_items.append(str(item))
            else:
                # Else, add to std_items
                std_items.append(item)

        # Call super method on bool_items
        super().addItems(bool_items)

        # Add a separator
        self.insertSeparator(len(bool_items))

        # Call super method on std_items
        super().addItems(std_items)

    # Add special get_box_value to convert bool-strings to bools
    def get_box_value(self, *value_sig):
        # Use normal method
        value = get_box_value(self, *value_sig, no_custom=True)

        # If the returned value is a bool and bools are used, convert value
        if value in ('True', 'False', 'None') and self.uses_bools:
            value = literal_eval(value)

        # Return value
        return(value)

    # Add special set_box_value to convert bools to bool-strings
    def set_box_value(self, value, *value_sig):
        # If the given value is a bool and bools are used, convert value
        if value in (True, False, None) and self.uses_bools:
            value = str(value)

        # Set value
        set_box_value(self, value, *value_sig, no_custom=True)


# Create function that returns spinbox to use for setting n_bins
def get_n_bins_box():
    # Make spinbox for bin count
    n_bins_box = GW.QSpinBox()
    n_bins_box.setRange(0, 100)
    n_bins_box.setSpecialValueText('auto')

    # Return it
    return(n_bins_box)


# Create custom spinbox that only allows values between 0 and 1
class UnitySpinBox(GW.QDoubleSpinBox):
    def __init__(self, *args, **kwargs):
        # Call super constructor
        super().__init__(*args, **kwargs)

        # Set unity range
        self.setDecimals(5)
        self.setRange(0, 1)


# %% CLASS DEFINITIONS
# Define config page for setting MPL configurations
class MPLConfigPage(GP.PluginConfigPage):
    # Define class attributes
    NAME = 'MPL'

    # This function sets up the MPL config page
    def init(self):
        # Create layout
        layout = GL.QVBoxLayout(self)

        # COLORMAPS
        # Create 'Colormaps' group box
        colormaps_group = GW.QGroupBox('Import colormaps')
        layout.addWidget(colormaps_group)
        colormaps_layout = GL.QFormLayout(colormaps_group)

        # Add label with explaining text
        cmaps_label = GW.QLabel(
            "You can import several Python packages that register colormaps in"
            " <u>matplotlib</u> before a figure is created. Packages that are "
            "not installed are ignored. Please provide each package separated "
            "by a comma, like: <i>cmasher, cmocean</i>")
        colormaps_layout.addRow(cmaps_label)

        # Create box for setting the colormap packages that must be imported
        cmap_pkg_box = GW.QLineEdit()
        colormaps_layout.addRow("Packages:", cmap_pkg_box)
        self.add_config_entry('cmap_packages', cmap_pkg_box, True)

        # RCPARAMS
        # Create 'rcParams' group box
        rcParams_group = GW.QGroupBox('rcParams')
        layout.addWidget(rcParams_group)
        rcParams_layout = GL.QVBoxLayout(rcParams_group)

        # Add label with explaining text
        rcParams_label = GW.QLabel(
            "You can modify most of the <u>rcParams</u> used in "
            "<u>matplotlib</u> below. Adding an entry will override the value "
            "for that specific parameter for every figure created.<br>"
            "See <a href=\"%s\">here</a> for an overview with descriptions of "
            "each parameter." %
            ("https://matplotlib.org/tutorials/introductory/customizing.html"))
        rcParams_layout.addWidget(rcParams_label)

        # Add separator
        rcParams_layout.addSeparator()

        # Create entries_box
        entries_box = GW.EntriesBox()
        rcParams_layout.addWidget(entries_box)
        self.add_config_entry('rcParams', entries_box)

        # Create list of all prefixes that should be skipped
        prefix_skip = ('_internal', 'agg', 'animation', 'backend',
                       'boxplot.bootstrap', 'datapath', 'docstring',
                       'figure.figsize', 'figure.max_open_warning',
                       'image.aspect', 'image.composite_image', 'image.origin',
                       'interactive', 'keymap', 'lines.color', 'mpl_toolkits',
                       'path', 'pdf', 'pgf', 'ps', 'savefig', 'svg',
                       'text.kerning_factor', 'tk', 'toolbar', 'verbose',
                       'webagg')

        # Create a combobox factory for text weights
        text_weight_box = create_combobox(
            ['normal', 'bold', 'bolder', 'lighter', '100', '200', '300', '400',
             '500', '600', '700', '800', '900'])

        # Create dict of all rcParams that use specific widgets
        key_widgets = {
            'axes.autolimit_mode': create_combobox(['data', 'round_numbers']),
            'axes.axisbelow': lambda: BoolComboBox([True, 'line', False]),
            'axes.edgecolor': GW.ColorBox,
            'axes.facecolor': GW.ColorBox,
            'axes.formatter.limits': lambda: GW.ItemsBox([int]*2),
            'axes.formatter.min_exponent': GW.QSpinBox,
            'axes.formatter.offset_threshold': GW.QSpinBox,
            'axes.formatter.use_locale': GW.QCheckBox,
            'axes.formatter.use_mathtext': GW.QCheckBox,
            'axes.formatter.useoffset': GW.QCheckBox,
            'axes.grid': GW.QCheckBox,
            'axes.grid.axes': create_combobox(['major', 'minor', 'both']),
            'axes.grid.which': create_combobox(['major', 'minor', 'both']),
            'axes.labelcolor': GW.ColorBox,
            'axes.labelpad': GW.FontSizeBox,
            'axes.labelsize': GW.FontSizeBox,
            'axes.labelweight': text_weight_box,
            'axes.linewidth': GW.FontSizeBox,
            'axes.prop_cycle': lambda: GW.GenericItemsBox(
                lambda: GW.ColorBox(False)),
            'axes.spines.bottom': GW.QCheckBox,
            'axes.spines.left': GW.QCheckBox,
            'axes.spines.right': GW.QCheckBox,
            'axes.spines.top': GW.QCheckBox,
            'axes.titlecolor': lambda: AutoToggleBox(GW.ColorBox(), 'auto'),
            'axes.titlelocation': create_combobox(['left', 'center', 'right']),
            'axes.titlepad': GW.FontSizeBox,
            'axes.titlesize': GW.FontSizeBox,
            'axes.titleweight': text_weight_box,
            'axes.unicode_minus': GW.QCheckBox,
            'axes.xmargin': UnitySpinBox,
            'axes.ymargin': UnitySpinBox,
            'axes3d.grid': GW.QCheckBox,
            'boxplot.boxprops.color': GW.ColorBox,
            'boxplot.boxprops.linestyle': GW.LineStyleBox,
            'boxplot.boxprops.linewidth': GW.FontSizeBox,
            'boxplot.capprops.color': GW.ColorBox,
            'boxplot.capprops.linestyle': GW.LineStyleBox,
            'boxplot.capprops.linewidth': GW.FontSizeBox,
            'boxplot.flierprops.color': GW.ColorBox,
            'boxplot.flierprops.linestyle': GW.LineStyleBox,
            'boxplot.flierprops.linewidth': GW.FontSizeBox,
            'boxplot.flierprops.marker': GW.MarkerStyleBox,
            'boxplot.flierprops.markeredgecolor': lambda: AutoToggleBox(
                GW.ColorBox(), 'none'),
            'boxplot.flierprops.markeredgewidth': GW.FontSizeBox,
            'boxplot.flierprops.markerfacecolor': lambda: AutoToggleBox(
                GW.ColorBox(), 'none'),
            'boxplot.flierprops.markersize': GW.FontSizeBox,
            'boxplot.meanline': GW.QCheckBox,
            'boxplot.meanprops.color': GW.ColorBox,
            'boxplot.meanprops.linestyle': GW.LineStyleBox,
            'boxplot.meanprops.linewidth': GW.FontSizeBox,
            'boxplot.meanprops.marker': GW.MarkerStyleBox,
            'boxplot.meanprops.markeredgecolor': lambda: AutoToggleBox(
                GW.ColorBox(), 'none'),
            'boxplot.meanprops.markerfacecolor': lambda: AutoToggleBox(
                GW.ColorBox(), 'none'),
            'boxplot.meanprops.markersize': GW.FontSizeBox,
            'boxplot.medianprops.color': GW.ColorBox,
            'boxplot.medianprops.linestyle': GW.LineStyleBox,
            'boxplot.medianprops.linewidth': GW.FontSizeBox,
            'boxplot.notch': GW.QCheckBox,
            'boxplot.showbox': GW.QCheckBox,
            'boxplot.showcaps': GW.QCheckBox,
            'boxplot.showfliers': GW.QCheckBox,
            'boxplot.showmeans': GW.QCheckBox,
            'boxplot.vertical': GW.QCheckBox,
            'boxplot.whiskerprops.color': GW.ColorBox,
            'boxplot.whiskerprops.linestyle': GW.LineStyleBox,
            'boxplot.whiskerprops.linewidth': GW.FontSizeBox,
            'boxplot.whiskers': GW.FloatLineEdit,
            'contour.corner_mask': lambda: BoolComboBox(
                [True, False, 'legacy']),
            'contour.negative_linestyle': GW.LineStyleBox,
            'date.autoformatter.day': GW.QLineEdit,
            'date.autoformatter.hour': GW.QLineEdit,
            'date.autoformatter.microsecond': GW.QLineEdit,
            'date.autoformatter.minute': GW.QLineEdit,
            'date.autoformatter.month': GW.QLineEdit,
            'date.autoformatter.second': GW.QLineEdit,
            'date.autoformatter.year': GW.QLineEdit,
            'errorbar.capsize': GW.FloatLineEdit,
            'figure.autolayout': GW.QCheckBox,
            'figure.constrained_layout.h_pad': UnitySpinBox,
            'figure.constrained_layout.hspace': UnitySpinBox,
            'figure.constrained_layout.use': GW.QCheckBox,
            'figure.constrained_layout.w_pad': UnitySpinBox,
            'figure.constrained_layout.wspace': UnitySpinBox,
            'figure.dpi': GW.FloatLineEdit,
            'figure.edgecolor': GW.ColorBox,
            'figure.facecolor': GW.ColorBox,
            'figure.frameon': GW.QCheckBox,
            'figure.subplot.bottom': UnitySpinBox,
            'figure.subplot.hspace': UnitySpinBox,
            'figure.subplot.left': UnitySpinBox,
            'figure.subplot.right': UnitySpinBox,
            'figure.subplot.top': UnitySpinBox,
            'figure.subplot.wspace': UnitySpinBox,
            'figure.titlesize': GW.FontSizeBox,
            'figure.titleweight': text_weight_box,
            'font.cursive': lambda: GW.GenericItemsBox(GW.QFontComboBox),
            'font.family': lambda: GW.GenericItemsBox(GW.QFontComboBox),
            'font.fantasy': lambda: GW.GenericItemsBox(GW.QFontComboBox),
            'font.monospace': lambda: GW.GenericItemsBox(GW.QFontComboBox),
            'font.sans-serif': lambda: GW.GenericItemsBox(GW.QFontComboBox),
            'font.serif': lambda: GW.GenericItemsBox(GW.QFontComboBox),
            'font.size': GW.FontSizeBox,
            'font.stretch': create_combobox(
                ['ultra-condensed', 'extra-condensed', 'condensed',
                 'semi-condensed', 'normal', 'semi-expanded', 'expanded',
                 'extra-expanded', 'ultra-expanded', 'wider', 'narrower']),
            'font.style': create_combobox(['normal', 'italic', 'oblique']),
            'font.variant': create_combobox(['normal', 'small-caps']),
            'font.weight': text_weight_box,
            'grid.alpha': UnitySpinBox,
            'grid.color': GW.ColorBox,
            'grid.linestyle': GW.LineStyleBox,
            'grid.linewidth': GW.FontSizeBox,
            'hatch.color': GW.ColorBox,
            'hatch.linewidth': GW.FontSizeBox,
            'hist.bins': get_n_bins_box,
            'image.cmap': GW.ColorMapBox,
            'image.interpolation': create_combobox(
                ['none', 'antialiased', 'nearest', 'bilinear', 'bicubic',
                 'spline16', 'spline36', 'hanning', 'hamming', 'hermite',
                 'kaiser', 'quadric', 'catrom', 'gaussian', 'bessel',
                 'mitchell', 'sinc', 'lanczos']),
            'image.lut': GW.IntLineEdit,
            'image.resample': GW.QCheckBox,
            'legend.borderaxespad': GW.FontSizeBox,
            'legend.borderpad': GW.FontSizeBox,
            'legend.columnspacing': GW.FontSizeBox,
            'legend.edgecolor': GW.ColorBox,
            'legend.facecolor': lambda: AutoToggleBox(
                GW.ColorBox(), 'inherit'),
            'legend.fancybox': GW.QCheckBox,
            'legend.fontsize': GW.FontSizeBox,
            'legend.framealpha': UnitySpinBox,
            'legend.frameon': GW.QCheckBox,
            'legend.handleheight': GW.FontSizeBox,
            'legend.handlelength': GW.FontSizeBox,
            'legend.handletextpad': GW.FontSizeBox,
            'legend.labelspacing': GW.FontSizeBox,
            'legend.loc': create_combobox(mpl.legend.Legend.codes.keys()),
            'legend.markerscale': GW.FloatLineEdit,
            'legend.numpoints': GW.IntLineEdit,
            'legend.scatterpoints': GW.IntLineEdit,
            'legend.shadow': GW.QCheckBox,
            'legend.title_fontsize': lambda: AutoToggleBox(
                GW.FontSizeBox(), None),
            'lines.antialiased': GW.QCheckBox,
            'lines.dash_capstyle': create_combobox(
                ['butt', 'projecting', 'round']),
            'lines.dash_joinstyle': create_combobox(
                ['bevel', 'miter', 'round']),
            'lines.dashdot_pattern': lambda: GW.ItemsBox([float]*4),
            'lines.dashed_pattern': lambda: GW.ItemsBox([float]*2),
            'lines.dotted_pattern': lambda: GW.ItemsBox([float]*2),
            'lines.linestyle': GW.LineStyleBox,
            'lines.linewidth': GW.FontSizeBox,
            'lines.marker': GW.MarkerStyleBox,
            'lines.markeredgecolor': lambda: AutoToggleBox(
                GW.ColorBox(), 'auto'),
            'lines.markeredgewidth': GW.FontSizeBox,
            'lines.markerfacecolor': lambda: AutoToggleBox(
                GW.ColorBox(), 'auto'),
            'lines.markersize': GW.FontSizeBox,
            'lines.scale_dashes': GW.QCheckBox,
            'lines.solid_capstyle': create_combobox(
                ['butt', 'projecting', 'round']),
            'lines.solid_joinstyle': create_combobox(
                ['bevel', 'miter', 'round']),
            'markers.fillstyle': create_combobox(
                ['full', 'left', 'right', 'bottom', 'top', 'none']),
            'mathtext.bf': GW.QLineEdit,
            'mathtext.cal': GW.QLineEdit,
            'mathtext.default': GW.QLineEdit,
            'mathtext.fallback_to_cm': GW.QCheckBox,
            'mathtext.fontset': create_combobox(
                ['dejavusans', 'cm', 'stix', 'stixsans', 'custom']),
            'mathtext.it': GW.QLineEdit,
            'mathtext.rm': GW.QLineEdit,
            'mathtext.sf': GW.QLineEdit,
            'mathtext.tt': GW.QLineEdit,
            'patch.antialiased': GW.QCheckBox,
            'patch.edgecolor': GW.ColorBox,
            'patch.facecolor': GW.ColorBox,
            'patch.force_edgecolor': GW.QCheckBox,
            'patch.linewidth': GW.FontSizeBox,
            'polaraxes.grid': GW.QCheckBox,
            'scatter.edgecolors': lambda: AutoToggleBox(
                GW.ColorBox(), 'face'),
            'scatter.marker': GW.MarkerStyleBox,
            'text.antialiased': GW.QCheckBox,
            'text.color': GW.ColorBox,
            'text.hinting': create_combobox(
                ['either', 'native', 'auto', 'none']),
            'text.hinting_factor': GW.IntLineEdit,
            'text.latex.preamble': GW.QLineEdit,
            'text.latex.preview': GW.QCheckBox,
            'text.usetex': GW.QCheckBox,
            'timezone': GW.QLineEdit,
            'xtick.alignment': create_combobox(['left', 'center', 'right']),
            'xtick.bottom': GW.QCheckBox,
            'xtick.color': GW.ColorBox,
            'xtick.direction': create_combobox(['in', 'out', 'inout']),
            'xtick.labelbottom': GW.QCheckBox,
            'xtick.labelsize': GW.FontSizeBox,
            'xtick.labeltop': GW.QCheckBox,
            'xtick.major.bottom': GW.QCheckBox,
            'xtick.major.pad': GW.FontSizeBox,
            'xtick.major.size': GW.FontSizeBox,
            'xtick.major.top': GW.QCheckBox,
            'xtick.major.width': GW.FontSizeBox,
            'xtick.minor.bottom': GW.QCheckBox,
            'xtick.minor.pad': GW.FontSizeBox,
            'xtick.minor.size': GW.FontSizeBox,
            'xtick.minor.top': GW.QCheckBox,
            'xtick.minor.visible': GW.QCheckBox,
            'xtick.minor.width': GW.FontSizeBox,
            'xtick.top': GW.QCheckBox,
            'ytick.alignment': create_combobox(
                ['left_baseline', 'center_baseline', 'right_baseline']),
            'ytick.color': GW.ColorBox,
            'ytick.direction': create_combobox(['in', 'out', 'inout']),
            'ytick.labelleft': GW.QCheckBox,
            'ytick.labelright': GW.QCheckBox,
            'ytick.labelsize': GW.FontSizeBox,
            'ytick.left': GW.QCheckBox,
            'ytick.major.left': GW.QCheckBox,
            'ytick.major.pad': GW.FontSizeBox,
            'ytick.major.right': GW.QCheckBox,
            'ytick.major.size': GW.FontSizeBox,
            'ytick.major.width': GW.FontSizeBox,
            'ytick.minor.left': GW.QCheckBox,
            'ytick.minor.pad': GW.FontSizeBox,
            'ytick.minor.right': GW.QCheckBox,
            'ytick.minor.size': GW.FontSizeBox,
            'ytick.minor.visible': GW.QCheckBox,
            'ytick.minor.width': GW.FontSizeBox,
            'ytick.right': GW.QCheckBox}

        # Initialize empty dict of entry_types
        entry_types = sdict()

        # Loop over all rcParams and determine what widget to use
        for key, value in rcParamsDefault.items():
            # Check if key starts with anything in prefix_skip
            if key.startswith(prefix_skip):
                # If so, continue
                continue

            # Obtain proper box
            box = key_widgets.get(key, GW.LongGenericBox)

            # If key is 'axes.prop_cycle', obtain the proper value
            if(key == 'axes.prop_cycle'):
                value = value.by_key()['color']

            # Add box to entry_types dict with default value
            entry_types[key] = (box, value)

        # Add all rcParams entries to the box
        entries_box.addEntryTypes(entry_types)

    # This function parses and processes a config section, and returns it
    def decode_config(self, section_dict):
        # Initialize empty dict of parsed config values
        config_dict = sdict()

        # Decode all values in section_dict
        for key, value in section_dict.items():
            # Convert to Python object
            value = literal_eval(value)

            # Add to dict
            config_dict[key] = value

        # Return config_dict
        return(config_dict)

    # This function returns a dict containing the default config values
    def get_default_config(self):
        return({'rcParams': {},
                'cmap_packages': ("cmasher, cmocean, palettable, colorcet")})

    # This function returns its config section, as required by config parser
    def encode_config(self, config_dict):
        # Initialize empty dict of section config values
        section_dict = sdict()

        # Loop over all arguments in config and encode them in
        for key, value in config_dict.items():
            # Add to dict
            section_dict[key] = '{!r}'.format(value)

        # Return section_dict
        return(section_dict)

    # This function applies the currently stored config
    def apply_config(self, config_dict):
        # Try to import all colormap packages
        pkgs = config_dict['cmap_packages'].replace(' ', '').split(',')
        for pkg in pkgs:
            try:
                import_module(pkg)
            except ImportError:
                pass

        # Obtain rcParams from config_dict and make a copy
        rcParams_config = dict(config_dict['rcParams'])

        # Modify the axes.prop_cycle if it is present
        if 'axes.prop_cycle' in rcParams_config:
            rcParams_config['axes.prop_cycle'] =\
                cycler(color=rcParams_config['axes.prop_cycle'])

        # Update MPL's rcParams with the values stored in this config page
        rcParams.update(rcParamsDefault)
        rcParams.update(rcParams_config)
